

# xml_set_attribv_occ
def test_xml_set_attribv_occ():
    from aiida_fleur.tools.xml_util import xml_set_attribv_occ
    pass

# xml_set_first_attribv
def test_xml_set_first_attribv():
    from aiida_fleur.tools.xml_util import xml_set_first_attribv
    pass

# xml_set_all_attribv
def test_xml_set_all_attribv():
    from aiida_fleur.tools.xml_util import xml_set_all_attribv
    pass

# xml_set_text
def test_xml_set_text():
    from aiida_fleur.tools.xml_util import xml_set_text
    pass

# xml_set_all_text
def test_xml_set_all_text():
    from aiida_fleur.tools.xml_util import xml_set_all_text
    pass

# create_tag
def test_create_tag():
    from aiida_fleur.tools.xml_util import create_tag
    pass

# delete_att
def test_delete_att():
    from aiida_fleur.tools.xml_util import delete_att
    pass

# delete_tag
def test_delete_tag():
    from aiida_fleur.tools.xml_util import delete_tag
    pass

# replace_tag
def test_replace_tag():
    from aiida_fleur.tools.xml_util import replace_tag
    pass

# set_species
def test_set_species():
    from aiida_fleur.tools.xml_util import set_species
    pass

# change_atomgr_att
def test_change_atomgr_att():
    from aiida_fleur.tools.xml_util import change_atomgr_att
    pass

# add_num_to_att
def test_add_num_to_att():
    from aiida_fleur.tools.xml_util import add_num_to_att
    pass

# eval_xpath
def test_eval_xpath():
    from aiida_fleur.tools.xml_util import eval_xpath
    pass

# eval_xpath2
def test_eval_xpath2():
    from aiida_fleur.tools.xml_util import eval_xpath2
    pass

# eval_xpath3
def test_eval_xpath3():
    from aiida_fleur.tools.xml_util import eval_xpath3
    pass

# get_xml_attribute
def test_get_xml_attribute():
    from aiida_fleur.tools.xml_util import get_xml_attribute
    pass

# get_inpxml_file_structure
# IMPORTANT: Here we need thats that tell us when the plugin has to be maintained, i.e Know thing in the inp schema where changed
# Is there a way to test for not yet know attributes? i.e if the plugin is complete? Back compatible?
# I.e make for each Fleur schema file, complete inp.xml file version a test if the attributes exists.

# make a test_class

    


# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
This module, contains a method to merge parameterData nodes used by the FLEUR inpgen.
This might also be of interest for other all-ellectron codes
"""
# TODO this should be made an inline calculation or workfunction to
# keep the proverance!
# Shall we allow for a python dictionary also instead of forcing paramteraData?
# but then we can not keep the provenace...

from aiida.orm import Code, CalculationFactory, DataFactory
from aiida.orm import load_node
from aiida.orm.data.base import Bool
from aiida.work import workfunction as wf

ParameterData = DataFactory('parameter')

def merge_parameter(ParameterData1, ParameterData2, overwrite=True):
    """
    Merges two parameterData nodes.
    Additive: uses all namelists of both.
    If they have a namelist in common. ParameterData2 will overwrite the namelist
    of parameterData. If this is not wanted. set overwrite = False.
    Then attributes of both will be added, but attributes from ParameterData1 won't
    be oeverwritten.


    param: AiiDA ParamterData Node
    param: AiiDA ParamterData Node

    returns: AiiDA ParamterData Node
    """

    # layout:
    # check input
    # get dictionaries
    # merge dictionaries into new dictionary
    # create a new parameterData node

    new_dict = {}
    atoms_dict = {}
    atomlist = []
    if not isinstance(ParameterData1, ParameterData):
        raise InputValidationError("ParameterData1, must be of "
                                           "type ParameterData")
    if not isinstance(ParameterData2, ParameterData):
        raise InputValidationError("ParameterData2, must be of "
                                           "type ParameterData")
    dict1 = ParameterData1.get_dict()
    dict2 = ParameterData2.get_dict()

    for key in dict1.keys():
        if 'atom' in key:
            val = dict1.pop(key)
            atomlist.append(val)

    for key in dict2.keys():
        if 'atom' in key:
            val = dict2.pop(key)
            atomlist.append(val)


    # TODO do something on atom list,
    # we dont want doubles, check element and Id? Keep first ones?

    for i, atom in enumerate(atomlist):
        # TODO check for duplicates? what about
        key = 'atom{}'.format(i)
        atoms_dict[key] = atom

    # merge all namelists except atoms
    if overwrite:
        new_dict = dict1.copy()
        new_dict.update(dict2)
    else:
        # add second one later?
        new_dict = dict2.copy()
        new_dict.update(dict1)
        # TODO or do we want to merge the value dicts? usually does not make sense
        # for all keys...
    new_dict.update(atoms_dict)

    # be carefull with atom namelist


    return ParameterData(dict=new_dict)


def merge_parameters(ParameterDataList, overwrite=True):
    """
    Merge together all parameter nodes in the given list.
    """
    paremeter_data_new = ParameterData(dict= {})

    for i, parameter in enumerate(ParameterDataList):
        if isinstance(parameter, ParameterData):
            # merge
            paremeter_data_new = merge_parameter(paremeter_data_new, parameter, overwrite=overwrite)
        else:
            print('WARNING: Entry : {} {} is not of type ParameterData, I skip it.'.format(i, parameter))

    return paremeter_data_new

@wf
def merge_parameter_wf(ParameterData1, ParameterData2, overwrite=Bool(True)):
    """
    workfunction of merge_parameters
    """
    paremeter_data_new = merge_parameter(ParameterData1, ParameterData2, overwrite=overwrite)

    return paremeter_data_new

'''
# TODO how to deal with a list? *args, prob is not the best, also it is not working here.
# makeing a methds m(self, *args, **kwargs) and setting some fallbacks, does not work, because self, cannot be parsed
# I guess...
@wf
def merge_parameters_wf(*ParameterDatas, overwrite=Bool(True)):
    """
    workfunction of merge_parameters
    """
    ParameterDataList = []
    for parameter in ParameterDatas:
        ParameterDataList.append(parameter)
    paremeter_data_new = merge_parameters(ParameterDataList, overwrite=overwrite)

    return paremeter_data_new
'''
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Merge a parameterData node.')
    parser.add_argument('--para1', type=ParameterData, dest='para1',
                        help='The first ParameterData node', required=True)
    parser.add_argument('--para2', type=ParameterData, dest='para2',
                        help='The second ParameterData node', required=True)
    parser.add_argument('--overwrite', type=bool, dest='overwrite',
                        help='Shall values given in ParameterData2 overwrite the values from the first ParameterData?', required=False)
    args = parser.parse_args()
    merge_parameter(paremeter_data_new=args.para1, parameter=args.para1, overwrite=args.overwrite)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################


"""
This is the worklfow 'corelevel' using the Fleur code, which calculates Binding
energies and corelevel shifts with different methods.
'divide and conquer'
"""
# TODO alow certain kpoint path, or kpoint node, so far auto


import os.path
from aiida.orm import Code, DataFactory
from aiida.work.workchain import WorkChain
from aiida.work.run import submit
from aiida.work.workchain import ToContext
#from aiida.work.process_registry import ProcessRegistry

from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida.work.workchain import while_, if_
from aiida_fleur.tools.create_corehole import create_corehole

StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')
RemoteData = DataFactory('remote')
FleurinpData = DataFactory('fleur.fleurinp')
FleurProcess = FleurCalculation.process()


class fleur_corelevel_wc(WorkChain):
    '''
    Turn key solution for the calculation of core level shift and Binding energies


    '''
    # wf_Parameters: ParameterData,
    '''
    'method' : ['initial', 'full_valence ch', 'half_valence_ch', 'ch', ...]
    'Bes' : [W4f, Be1s]
    'CLS' : [W4f, Be1s]
    'atoms' : ['all', 'postions' : []]
    'references' : ['calculate', or
    'scf_para' : {...}, 'default'
    'relax' : True
    'relax_mode': ['Fleur', 'QE Fleur', 'QE']
    'relax_para' : {...}, 'default'
    'calculate_doses' : False
    'dos_para' : {...}, 'default'
    '''
    '''
    # defaults
    default wf_Parameters::
    'method' : 'initial'
    'atoms' : 'all
    'references' : 'calculate'
    'scf_para' : 'default'
    'relax' : True
    'relax_mode': 'QE Fleur'
    'relax_para' : 'default'
    'calculate_doses' : False
    'dos_para' : 'default'
    '''

    _workflowversion = "0.0.1"

    @classmethod
    def define(cls, spec):
        super(corelevel, cls).define(spec)
        spec.input("wf_parameters", valid_type=ParameterData, required=False,
                   default=ParameterData(dict={
                                            'method' : 'initial',
                                            'atoms' : 'all',
                                            'references' : 'calculate',
                                            'calculate_doses' : False,
                                            'relax' : True,
                                            'relax_mode': 'QE Fleur',
                                            'relax_para' : 'default',
                                            'scf_para' : 'default',
                                            'dos_para' : 'default'}))
        spec.input("fleurinp", valid_type=FleurinpData, required=True)
        spec.input("fleur", valid_type=Code, required=True)
        spec.input("structure", valid_type=StructureData, required=False)
        spec.input("calc_parameters", valid_type=ParameterData, required=False)
        spec.outline(
            cls.check_input,
            if_(cls.initalstate)(
                cls.calculate_inital),
            cls.create_new_fleurinp,
            cls.run_fleur,
            cls.run_scfs,
            cls.collect_results,
            cls.return_results
        )
        #spec.dynamic_output()


    def start(self):
        '''
        check parameters, what condictions? complete?
        check input nodes
        '''
        ### input check ### ? or done automaticly, how optional?
        # check if fleuinp corresponds to fleur_calc
        print('started bands workflow version {}'.format(self._workflowversion))
        print("Workchain node identifiers: ")#{}"
              #"".format(ProcessRegistry().current_calc_node))



    def create_new_fleurinp(self):
        """
        create a new fleurinp from the old with certain parameters
        """
        # TODO allow change of kpoint mesh?, tria?
        wf_dict = self.inputs.wf_parameters.get_dict()
        nkpts = wf_dict.get('nkpts', 500)
        # how can the user say he want to use the given kpoint mesh, ZZ nkpts : False/0
        sigma = wf_dict.get('sigma', 0.005)
        emin = wf_dict.get('emin', -0.30)
        emax = wf_dict.get('emax', 0.80)

        fleurmode = FleurinpModifier(self.inputs.fleurinp)

        #change_dict = {'band': True, 'ndir' : -1, 'minEnergy' : self.inputs.wf_parameters.get_dict().get('minEnergy', -0.30000000),
        #'maxEnergy' :  self.inputs.wf_parameters.get_dict().get('manEnergy','0.80000000'),
        #'sigma' :  self.inputs.wf_parameters.get_dict().get('sigma', '0.00500000')}
        change_dict = {'band': True, 'ndir' : 0, 'minEnergy' : emin,
                       'maxEnergy' : emax, 'sigma' : sigma} #'ndir' : 1, 'pot8' : True

        fleurmode.set_inpchanges(change_dict)

        if nkpts:
            fleurmode.set_nkpts(count=nkpts)
            #fleurinp_new.replace_tag()

        fleurmode.show(validate=True, display=False) # needed?
        fleurinp_new = fleurmode.freeze()
        self.ctx.fleurinp1 = fleurinp_new
        print(fleurinp_new)
        #print(fleurinp_new.folder.get_subfolder('path').get_abs_path(''))

    def get_inputs_fleur(self):
        '''
        get the input for a FLEUR calc
        '''
        inputs = FleurProcess.get_inputs_template()

        fleurin = self.ctx.fleurinp1
        #print fleurin
        remote = self.inputs.remote
        inputs.parent_folder = remote
        inputs.code = self.inputs.fleur
        inputs.fleurinpdata = fleurin

        # TODO nkpoints decide n core

        core = 12 # get from computer nodes per machine
        inputs._options.resources = {"num_machines": 1, "num_mpiprocs_per_machine" : core}
        inputs._options.max_wallclock_seconds = 30 * 60

        if self.ctx.serial:
            inputs._options.withmpi = False # for now
            inputs._options.resources = {"num_machines": 1}

        if self.ctx.queue:
            inputs._options.queue_name = self.ctx.queue
            print(self.ctx.queue)
        # if code local use
        #if self.inputs.fleur.is_local():
        #    inputs._options.computer = computer
        #    #else use computer from code.
        #else:
        #    inputs._options.queue_name = 'th1'

        if self.ctx.serial:
            inputs._options.withmpi = False # for now
            inputs._options.resources = {"num_machines": 1}

        return inputs

    def run_fleur(self):
        '''
        run a fleur calculation
        '''
        FleurProcess = FleurCalculation.process()
        inputs = {}
        inputs = self.get_inputs_fleur()
        #print inputs
        future = submit(FleurProcess, **inputs)
        print('run Fleur in band workflow')

        return ToContext(last_calc=future)

    def return_results(self):
        '''
        return the results of the calculations
        '''
        # TODO more here
        print('Band workflow Done')
        print('A bandstructure was calculated for fleurinpdata {} and is found under pk={}, '
              'calculation {}'.format(self.inputs.fleurinp, self.ctx.last_calc.pk, self.ctx.last_calc))

        #check if band file exists: if not succesful = False
        #TODO be careful with general bands.X

        bandfilename = 'bands.1' # ['bands.1', 'bands.2', ...]
        # TODO this should be easier...
        last_calc_retrieved = self.ctx.last_calc.get_outputs_dict()['retrieved'].folder.get_subfolder('path')
        bandfilepath = self.ctx.last_calc.get_outputs_dict()['retrieved'].folder.get_subfolder('path').get_abs_path(bandfilename)
        print(bandfilepath)
        #bandfilepath = "path to bandfile" # Array?
        if os.path.isfile(bandfilepath):
            self.ctx.successful = True
        else:
            bandfilepath = None
            print('!NO bandstructure file was found, something went wrong!')
        #TODO corret efermi:
        # get efermi from last calculation
        efermi1 = self.inputs.remote.get_inputs()[-1].res.fermi_energy
        #get efermi from this caclulation
        efermi2 = self.ctx.last_calc.res.fermi_energy
        diff_efermi = efermi1 - efermi2
        # store difference in output node
        # adjust difference in band.gnu
        #filename = 'gnutest2'

        outputnode_dict ={}

        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['Warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['diff_efermi'] = diff_efermi
        #outputnode_dict['last_calc_pk'] = self.ctx.last_calc.pk
        #outputnode_dict['last_calc_uuid'] = self.ctx.last_calc.uuid
        outputnode_dict['bandfile'] = bandfilepath
        outputnode_dict['last_calc_uuid'] = self.ctx.last_calc.uuid
        outputnode_dict['last_calc_retrieved'] = last_calc_retrieved
        #print outputnode_dict
        outputnode = ParameterData(dict=outputnode_dict)
        outdict = {}
        outdict['band_out'] = outputnode
        #print outdict
        for k, v in outdict.iteritems():
            self.out(k, v)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the parser for a inpgen run (calculation) and methods for
parsing different files produced by inpgen, which form the input for the FLEUR
code.

Please implement file parsing routines that they can be executed from outside
the parser. Makes testing and portability easier. Also without using aiida_classes,
that they might be useful to external tools
"""
#TODO: maybe something from the out files should be saved in the db

import os
from aiida.parsers.parser import Parser
from aiida_fleur.data.fleurinp import FleurinpData
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation
from aiida_fleur.parsers import FleurOutputParsingError


class Fleur_inputgenParser(Parser):
    """
    This class is the implementation of the Parser class for the FLEUR inpgen.
    It takes the files recieved from an inpgen calculation and creates AiiDA
    nodes for the Database. From the inp.xml file a FleurinpData object is
    created, also some information from the out file is stored in a
    ParameterData node.
    """

    _setting_key = 'parser_options'

    def __init__(self, calc):
        """
        Initialize the instance of Fleur_inputgenParser
        """
        # check for valid input
        if not isinstance(calc, FleurinputgenCalculation):
            raise FleurOutputParsingError(
                "Input calc must be a FleurInpgenCalculation")

        # these files should be at least present after success of inpgen
        self._default_files = {calc._OUTPUT_FILE_NAME, calc._INPXML_FILE_NAME}
        self._other_files = {calc._SHELLOUT_FILE_NAME}
        #"enpara","inp","sym.out", "fort.93","struct.xsf"}
        #plus other special files? corelevels.xx, ... get from calc object
        super(Fleur_inputgenParser, self).__init__(calc)

    def parse_with_retrieved(self, retrieved):
        """
        Receives as input a dictionary of the retrieved nodes from an inpgen run.
        Does all the logic here.

        :return: a dictionary of AiiDA nodes for storing in the database.
        """

        successful = True
        has_xml_inpfile = False
        #has_error = False

        # select the folder object
        # Check that the retrieved folder is there
        try:
            out_folder = retrieved[self._calc._get_linkname_retrieved()]
        except KeyError:
            self.logger.error("No retrieved folder found")
            return False, ()

        # check what is inside the folder
        list_of_files = out_folder.get_folder_list()
        self.logger.info("file list {}".format(list_of_files))

        if self._calc._INPXML_FILE_NAME not in list_of_files:
            successful = False
            self.logger.error(
                "XML inp not found '{}'".format(self._calc._INPXML_FILE_NAME))
        else:
            has_xml_inpfile = True

        for file1 in self._default_files:
            if file1 not in list_of_files:
                successful = False
                self.logger.warning(
                    "'{}' file not found in retrived folder, it was probable "
                    "not created by inpgen".format(file1))
        # TODO what about other files?

        # TODO parse out file of inpgen
        #output_data = ParameterData(dict=out_dict)
        #link_name = self.get_linkname_outparams()
        #new_nodes_list = [(link_name, output_data)]
        #return successful,new_nodes_list

        new_nodes_list = []
        if self._calc._ERROR_FILE_NAME in list_of_files:
            errorfile = os.path.join(out_folder.get_abs_path('.'),
                                     self._calc._ERROR_FILE_NAME)
            # read
            error_file_lines = ''
            try:
                with open(errorfile, 'r') as efile:
                    error_file_lines = efile.read()# Note: read(),not readlines()
            except IOError:
                self.logger.error(
                    "Failed to open error file: {}.".format(errorfile))
            # if not empty, has_error equals True, parse error.
            if error_file_lines:
                self.logger.error(
                    "The following was written to the error file {} : \n '{}'"
                    "".format(self._calc._ERROR_FILE_NAME, error_file_lines))
                #has_error = True
                successful = False
                return successful, ()

        if has_xml_inpfile:
            # read xmlinp file into an etree
            inpxmlfile = os.path.join(out_folder.get_abs_path('.'),
                                      self._calc._INPXML_FILE_NAME)
            #tree = etree.parse(inpxmlfile)
            #root = tree.getroot()

            # convert etree into python dictionary
            #inpxml_params_dict = inpxml_todict(tree.getroot(), inpxmlstructure)
            #self.inpxmlstructure)#tags_several)#.xpath('/fleurInput')[0])

            # convert the dictionary into an AiiDA object
            #Some values, errors and co?
            #inpxml_params = ParameterData(dict=inpxml_params_dict)
            #link_name = self.get_linkname_outparams()
            # this will be accesible with res
            #new_nodes_list.append((link_name, inpxml_params))

            #check if inpgen was invokes with other options
            fleurinp_data = FleurinpData(files=[inpxmlfile])
            # if kpoints
            # fleurinp_data.set_kpoints
            #, symoutfile, enparafile])
            self.logger.info('FleurinpData initialized')
            #self.logger.info
            link_name_fleurinp = 'fleurinpData'
            # return it to the execmanager
            new_nodes_list.append((link_name_fleurinp, fleurinp_data))

        return successful, new_nodes_list




import pytest
#from aiida.utils.fixtures import fixture_manager

@pytest.fixture(scope='session')
def aiida_profile():
    from aiida.utils.fixtures import fixture_manager
    
    with fixture_manager() as fixture_mgr:
        fixture_mgr.create_profile()
        yield fixture_mg
        

#@pytest.fixture(scope='function')
#def test_data(aiida_profile):
#    # load my test data
#    yield
#    aiida_profile.reset_db()

##def test_my_stuff(test_data):
#   # run a test
#   print('test_my_stuf works')



import pytest

# get_natoms_element
def test_get_natoms_element_Be2W():
    from aiida_fleur.tools.common_fleur_wf_util import get_natoms_element
    
    assert get_natoms_element('Be2W') == {'Be' : 2, 'W' : 1}

# get_atomprocent
def test_get_atomprocent_Be24W2():
    from aiida_fleur.tools.common_fleur_wf_util import get_atomprocent
    assert get_atomprocent('Be24W2') == {'Be': 24./26., 'W' : 2./26.}


#TODO
# get_weight_procent
def test_get_weight_procent():
    from aiida_fleur.tools.common_fleur_wf_util import get_weight_procent
    pass

# determine_formation_energy
def test_determine_formation_energy():
    from aiida_fleur.tools.common_fleur_wf_util import determine_formation_energy
    
    # form energy is per atom here...
    form_en_exp = [-0.16666666666666666, 0.0]
    form_en_dict_exp = {'Be2W': -0.16666666666666666, 'BeW': 0.0}
    form_en, form_en_dict = determine_formation_energy({'BeW' : 2, 'Be2W' : 2.5}, {'Be' : 1, 'W' : 1})
    assert form_en == form_en_exp
    assert form_en_dict == form_en_dict_exp


# inpgen_dict_set_mesh
def test_inpgen_dict_set_mesh():
    from aiida_fleur.tools.common_fleur_wf_util import inpgen_dict_set_mesh
    pass

#inpgen_dict_set_mesh(Be_para.get_dict(), mesh)

# powerset
def test_powerset():
    from aiida_fleur.tools.common_fleur_wf_util import powerset
    
    res = [(), ('Be',), ('W',), ('Be2W',), ('Be', 'W'), ('Be', 'Be2W'), ('W', 'Be2W'), ('Be', 'W', 'Be2W')]
    length = len(res)
    assert powerset([1, 2, 3]) == [(), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]
    assert powerset(['Be', 'W', 'Be2W']) == res
    assert length == 8

# determine_reactions
def test_determine_reactions():
    from aiida_fleur.tools.common_fleur_wf_util import determine_reactions

    res = ['1*Be12W->1*Be12W', '1*Be12W->1*Be2W+10*Be',
           '2*Be12W->1*Be2W+1*Be22W', '1*Be12W->12*Be+1*W',
           '11*Be12W->5*W+6*Be22W']
    n_equations = len(res)

    assert determine_reactions('Be12W', ['Be12W', 'Be2W', 'Be', 'W', 'Be22W']) == res
    assert n_equations == 5

# convert_eq_to_dict
def test_convert_eq_to_dict():
    from aiida_fleur.tools.common_fleur_wf_util import convert_eq_to_dict
    
    res_dict = {'products': {'Be': 15, 'Be2Ti': 1}, 'educts': {'Be12Ti': 1}}
    assert convert_eq_to_dict('1*Be12Ti->10*Be+1*Be2Ti+5*Be') == res_dict


# get_enhalpy_of_equation
def test_get_enhalpy_of_equation():
    from aiida_fleur.tools.common_fleur_wf_util import get_enhalpy_of_equation
    pass    

# balance_equation
def test_balance_equation():
    from aiida_fleur.tools.common_fleur_wf_util import balance_equation
    pass

# test
#print(balance_equation("C7H16+O2 -> CO2+H2O"))
#print balance_equation("Be12W->Be2W+W+Be")#+Be12W+Be+Be22W")
#print balance_equation("Be12WO->Be2WO+W+Be+O2")#+Be12W+Be+Be22W")
#print balance_equation("Be12W->Be22W+Be12W")
#print balance_equation("Be12W->Be12W")

#1*C7H16+11*O2 ->7* CO2+8*H2O
#None
#None
#None
#1*Be12W->1*Be12W







# dict_merger
def test_dict_merger_interface_dicts_lists_str_int():
    from aiida_fleur.tools.ParameterData_util import dict_merger

    dict1 = {'a1' : {'b1' : [1]}}
    dict2 = {'a1' : {'b1': [2,3]}}
    assert dict_merger(dict1, dict2) == {'a1' : {'b1' : [1,2,3]}}

    dict3 = {'a1' : {'b2' : 1}}
    dict4 = {'a1' : {'b2' : 1}}
    assert dict_merger(dict3, dict4) == {'a1' : {'b2' : 2}}
    
    dict5 = {'a1' : {'b3' : 'a'}}
    dict6 = {'a1' : {'b3' : 'b' }, 'a2': [1]}
    assert dict_merger(dict5, dict6) == {'a1' : {'b3' : 'ab'}, 'a2' : [1]}


# extract_elementpara
def test_extract_elementpara_interface_W():
    from aiida_fleur.tools.ParameterData_util import extract_elementpara

    para_dict = {'a' : 1, 'atom' : {'element' : 'H', 'rmt': 1}, 'atom1' : {'element' : 'W', 'rmt' : 4}}
    assert extract_elementpara(para_dict, 'W') == {'a' : 1, 'atom1' : {'element' : 'W', 'rmt' : 4}}




# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
Input plug-in for a FLEUR ciculation. fleur.x
"""
# TODO:
# polishing
# think about exception. warning policy.
# TODO maybe allow only single file names not *
# TODO maybe check the settings key values, make a list of all fleur files?
import os
#from lxml import etree
#from lxml.etree import XMLSyntaxError

from aiida.orm.calculation.job import JobCalculation
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.remote import RemoteData
from aiida_fleur.data.fleurinp import FleurinpData
from aiida.common.utils import classproperty
from aiida.common.exceptions import InputValidationError, ValidationError
from aiida.common.exceptions import UniquenessError



class FleurCalculation(JobCalculation):
    """
    Main DFT code of the FLEUR code
    For more information about the FLEUR-code family, go to http://www.flapw.de/
    """
    def _init_internal_params(self):
        super(FleurCalculation, self)._init_internal_params()

        ######### Only this should be to be maintained! #########

        # Default fleur output parser
        self._default_parser = 'fleur.fleurparser'
        #self._default_parser = 'fleur.fleur'

        # should a kpt node be used or fleur generate the mesh?
        self._use_kpoints = False
        self._INPXML_FILE_NAME = 'inp.xml'

        # Default input and output files
        self._DEFAULT_INPUT_FILE = 'inp.xml'#self._set_INPXML_FILE_PATH()
        #self._FLEURINPDATA_FOLDER
        #fleurinp.get_file_abs_path(self._INPXML_FILE_NAME)
        #'inp.xml' # this has to change the file is under FleurinpData
        self._DEFAULT_OUTPUT_FILE = 'out.xml'

        # Name of all files in FLEUR
        # TODO think also to oursource this in a FleurCalc structure, and add
        # as info when what file has to be copied

        # these will be shown in AiiDA
        self._OUTPUT_FILE_NAME = 'aiida.out' # Shell output
        self._INPUT_FILE_NAME = 'inp.xml'
        #fleur file names:

        # needed for calc

        self._OUTXML_FILE_NAME = 'out.xml'
        self._INP_FILE_NAME = 'inp'
        self._ENPARA_FILE_NAME = 'enpara'
        self._SYMOUT_FILE_NAME = 'sym.out'
        self._CDN1_FILE_NAME = 'cdn1'
        self._SHELLOUTPUT_FILE_NAME = 'shell.out'
        self._ERROR_FILE_NAME = 'out.error'
         # other
        self._OUT_FILE_NAME = 'out'
        self._CDNC_FILE_NAME = 'cdnc' # core charge density
        self._TIME_INFO_FILE_NAME = 'time.info'
        self._KPTS_FILE_NAME = 'kpts'
        self._QPTS_FILE_NAME = 'qpts'
        self._PLOT_INP_FILE_NAME = 'plot_inp'
        self._BROYD_FILE_NAME = 'broyd*'
        self._POT_FILE_NAME = 'pot*'
        self._POT1_FILE_NAME = 'pottot'
        self._POT2_FILE_NAME = 'potcoul'
        self._STRUCTURE_FILE_NAME = 'struct.xcf'
        self._STARS_FILE_NAME = 'stars'
        self._WKF2_FILE_NAME = 'wkf2'
        self._CDN_HDF5_FILE_NAME = 'cdn.hdf'
        self._CDN_LAST_HDF5_FILE_NAME = 'cdn_last.hdf'


        # special out files
        self._DOS_FILE_NAME = 'DOS.*'
        self._DOSINP_FILE_NAME = 'dosinp'
        self._BAND_GNU_FILE_NAME = 'band.gnu'
        self._BAND_FILE_NAME = 'bands.*'
        self._NEW_XMlINP_FILE_NAME = 'inp_new.xml'

        # helper files
        self._FLEUR_WARN_ONLY_INFO_FILE_NAME = 'FLEUR_WARN_ONLY'
        self._JUDFT_WARN_ONLY_INFO_FILE_NAME = 'JUDFT_WARN_ONLY'
        self._QFIX_FILE_NAME = 'qfix'


        # forces and relaxation files
        self._FORCE_FILE_NAME = 'forces.dat'

        # jij files
        self._JENERG_FILE_NAME = 'jenerg'
        self._MCINP_FILE_NAME = 'MCinp'
        self._QPTSINFO_FILE_NAME = 'qptsinfo'
        self._SHELL_FILE_NAME = 'shells'
        self._JCONST_FILE_NAME = 'jconst'

        # files for lda+U
        self._NMMPMAT_FILE_NAME = 'n_mmp_mat'

        # files for hybrid functionals
        self._COULOMB1_FILE_NAME = 'coulomb1'
        self._MIXBAS_FILE_NAME = 'mixbas'
        self._CMT_FIlE_NAME = 'cmt'
        self._CZ_FILE_NAME = 'cz'
        self._OLAP_FILE_NAME = 'olap'
        self._VR0_FILE_NAME = 'vr0'

        # files non-collinear calculation
        self._RHOMAT_INP_FILE_NAME = 'rhomat_inp'
        self._RHOMAT_OUT_FILE_NAME = 'rhomat_out'
        self._CDN_FILE_NAME = 'cdn'
        self._DIROFMAG_FILE_NAME = 'dirofmag'

        # files for Wannier 90
        self._W90KPTS_FILE_NAME = 'w90kpts'
        self._PROJ_FILE_NAME = 'proj'
        self._WANN_INP_FILE_NAME = 'wann_inp'
        self._BKPTS_FILE_NAME = 'bkpts'
        self._WFMMN_FILE_NAME = 'WF*.mmn'
        self._WFAMN_FILE_NAME = 'WF*.amn'
        self._WFWIN_FILE_NAME = 'WF*.win'
        self._WFWOUT_FILE_NAME = 'WF*.wout'
        self._UNK_FILE_NAME = 'UNK*'
        self._KPTSMAP_FILE_NAME = 'kptsmap'
        self._PROJGEN_INP_FILE_NAME = 'projgen_inp'
        self._IONS_FILE_NAME = 'IONS'
        self._POLARIZATION_OUT_FILE_NAME = 'polarization_out'
        self._HOPPING_FILE_NAME = 'hopping.*'
        self._WF1HSOMTX_FILE_NAME = 'WF1.hsomtx'
        self._RSSOCMAT_FILE_NAME = 'rssocmat.1'
        self._RSNABLA_FILE_NAME = 'rsnabla.*'
        self._WFNABL_FILE_NAME = 'WF*.nabl'

        # copy file lists. I rather dont like this.
        # Might gives rise to a lot of possible erros, if files or not there,
        #or Fleur did not created same, or at some point they will not be
        # deleted remotely.

        ########## Policy
        # we store everything needed for a further run in the local repository
        #(inp.xml, cdn1), also all important results files.
        # these will ALWAYS be copied from the local repository to the maschine
        # If a parent calculation exists, other files will be copied remotely
        #######

        #all possible files first chargedensity
        self._copy_filelist1 = []
        self._copy_filelist1 = [self._INP_FILE_NAME,
                                self._ENPARA_FILE_NAME,
                                self._SYMOUT_FILE_NAME,
                                self._CDN1_FILE_NAME,
                                self._KPTS_FILE_NAME,
                                self._STARS_FILE_NAME,
                                self._WKF2_FILE_NAME]

        #after inpgen, before first chargedensity
        self._copy_filelist_inpgen = [self._INPXML_FILE_NAME]

        #for after fleur SCF [name, detination_name]
        self._copy_filelist_scf1 = [[self._CDN1_FILE_NAME, self._CDN1_FILE_NAME]]
        self._copy_filelist_scf2 = [[self._CDN_LAST_HDF5_FILE_NAME, self._CDN_HDF5_FILE_NAME]]
        #self._INPXML_FILE_NAME, comes from fleurinpdata
        self._copy_filelist_scf = [[self._CDN1_FILE_NAME, self._CDN1_FILE_NAME],
                                   [self._INPXML_FILE_NAME, self._INPXML_FILE_NAME]]
        self._copy_filelist_scf2_1 = [[self._CDN_LAST_HDF5_FILE_NAME, self._CDN_HDF5_FILE_NAME],
                                   [self._INPXML_FILE_NAME, self._INPXML_FILE_NAME]]
        self._copy_filelist_scf_remote = [self._BROYD_FILE_NAME]
        self._copy_filelist3 = []
        self._copy_filelist3 = [self._INP_FILE_NAME,
                                self._ENPARA_FILE_NAME,
                                self._SYMOUT_FILE_NAME,
                                self._CDN1_FILE_NAME,
                                self._KPTS_FILE_NAME,
                                self._STARS_FILE_NAME,
                                self._WKF2_FILE_NAME,
                                self._BROYD_FILE_NAME,
                                self._OUT_FILE_NAME,
                                self._POT_FILE_NAME]

        #files need for rerun
        self._copy_filelist3 = []
        self._copy_filelist_dos = [self._INPXML_FILE_NAME,
                                   self._CDN1_FILE_NAME]
        self._copy_filelist_band = [self._INPXML_FILE_NAME,
                                    self._POT_FILE_NAME,
                                    self._CDN1_FILE_NAME]


        self._copy_filelist_hybrid = []
        self._copy_filelist_jij = []

        #possible settings_dict keys
        self._settings_keys = ['additional_retrieve_list', 'remove_from_retrieve_list',
                               'additional_remotecopy_list', 'remove_from_remotecopy_list'
                               'cmdline']
        #possible modes?
        self._fleur_modes = ['band', 'dos', 'forces', 'chargeDen',
                             'latticeCo', 'scf']
    #_DEFAULT_INPUT_FILE = 'inp.xml'
    @classproperty
    def _use_methods(cls):
        """
        Extend the parent _use_methods with further keys.
        """
        retdict = JobCalculation._use_methods
        retdict.update({
            "fleurinpdata": {
                'valid_types': FleurinpData,
                'additional_parameter': None,
                'linkname': 'fleurinpdata',
                'docstring': (
                    "Use a FleruinpData node that specifies the input parameters"
                    "usually copy from the parent calculation, basicly makes"
                    "the inp.xml file visible in the db and makes sure it has "
                    "the files needed."),
                },
            "parent_folder": {
                'valid_types': RemoteData,
                'additional_parameter': None,
                'linkname': 'parent_calc_folder',
                'docstring': (
                    "Use a remote or local repository folder as parent folder "
                    "(also for restarts and similar). It should contain all the "
                    "needed files for a Fleur calc, only edited files should be "
                    "uploaded from the repository."),
                },
            "settings": {
                'valid_types': ParameterData,
                'additional_parameter': None,
                'linkname': 'settings',
                'docstring': (
                    "This parameter data node is used to specify for some "
                    "advanced features how the plugin behaves. You can add files"
                    "the retrieve list, or add command line switches, "
                    "for all available features here check the documentation."),

            }})
            #
            #"parent_calc":{
            #   'valid_types': (FleurinputgenCalculation, FleurCalculation),
            #   'additional_parameter': None,
            #   'linkname': 'parent_calc',
            #   'docstring': ("Use a parent calculation from which to copy"
            #                 "files, and so on."),
            #
            #  },
            #"kpoints" : {
            #   'valid_types': KpointsData,
            #   'additional_parameter': None,
            #   'linkname': 'kpoints',
            #   'docstring': "Connect to k-point node from inpgen or other calc.",
            #   }


        return retdict

    @classproperty
    def _OUTPUT_FOLDER(cls):
        return './'
    '''
    @classproperty
    def _set_INPXML_FILE_PATH(self, cls, fleurinp=None):
        if fleruinp:
            if not isinstance(fleurinp, FleurinpData):
                self._DEFAULT_INPUT_FILE = 'inp.xml'
            else:
                self._DEFAULT_INPUT_FILE(
                    fleurinp.get_file_abs_path(self._INPXML_FILE_NAME))
        else:
            self._DEFAULT_INPUT_FILE = 'inp.xml'

    @classproperty
    def _FLEURINPDATA_FOLDER(self, cls, inputdict):
        try:
            fleurinp = inputdict[self.get_linkname('fleurinpdata')]
        except:
            return None
        if fleurinp is None:
            xml_inp_dict = {}
        else:
            if not isinstance(fleurinp, FleurinpData):
                raise InputValidationError(
                    "The FleurinpData node given is not of type FleurinpData.")
        return fleurinp.get_file_abs_path(self._INPXML_FILE_NAME)

    def set_inppath(cls):
        self._DEFAULT_INPUT_FILE = fleurinp.get_file_abs_path(
                                                         self._INPXML_FILE_NAME)
    '''
    def _prepare_for_submission(self, tempfolder, inputdict):
        """
        This is the routine to be called when you make a fleur calculation
        Here should be checked if all the files are there to run fleur.
        And input files (inp.xml) can be modified.

        :param tempfolder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputdata_dict (without the Code!)
        """

        # TODO how to check if code compiled with HDF5?
        # Idea: in description of code can be written wirh what libs the code was compiled,
        # and we check in the description for certain keywords... if we have the code node...
        # also ggf, to be back comportable, the plugin should know the version number...

        #   from aiida.common.utils import get_unique_filename, get_suggestion

        local_copy_list = []
        remote_copy_list = []
        remote_symlink_list = []
        mode_retrieved_filelist = []
        #filelocal_copy_list = []
        #filelist_tocopy = []
        filelist_tocopy_remote = []
        settings_dict = {}

        #fleur_calc = False
        #new_inp_file = False
        #ignore_mode = False
        has_fleurinp = False
        has_parent = False
        #restart_flag = False
        fleurinpgen = False
        copy_remotely = True
        with_hdf5 = False

        ##########################################
        ############# INPUT CHECK ################
        ##########################################

        try:
            code = inputdict.pop(self.get_linkname('code'))
        except KeyError:
            raise InputValidationError("No code specified for this calculation")

        codesdesc = code.description
        # TODO ggf also check settings
        if codesdesc is not None:
            if 'hdf5' in codesdesc: #or 'Hdf5' or 'HDF5') in codesdesc:
                with_hdf5 = True
            elif 'Hdf5' in codesdesc:
                with_hdf5 = True
            elif 'HDF5' in codesdesc:
                with_hdf5 = True
            else:
                with_hdf5 = False
        # a Fleur calc can be created from a fleurinpData alone
        #(then no parent is needed) all files are in the repo, but usually it is
        # a child of a inpgen calc or an other fleur calc (some or all files are
        # in a remote source). if the User has not changed something, the
        #calculation does not need theoretical a new FleurinpData it could use
        #the one from the parent, but the plug-in desgin is in a way that it has
        # to be there and it just copies files if changes occured..

        fleurinp = inputdict.pop(self.get_linkname('fleurinpdata'), None)
        if fleurinp is None:
            #xml_inp_dict = {}
            has_fleurinp = False
        else:
            if not isinstance(fleurinp, FleurinpData):
                raise InputValidationError(
                    "The FleurinpData node given is not of type FleurinpData.")
            has_fleurinp = True
        parent_calc_folder = inputdict.pop(self.get_linkname('parent_folder'),
                                           None)

        if parent_calc_folder is None:
            has_parent = False
            if not has_fleurinp:
                raise InputValidationError(
                    "No parent calculation found and no fleurinp data "
                    "given, need either one or both for a "
                    "fleurcalculation.")
        else: #
            if not isinstance(parent_calc_folder, RemoteData):
                raise InputValidationError("parent_calc_folder, if specified,"
                                           "must be of type RemoteData")

            # extract parent calculation
            parent_calcs = parent_calc_folder.get_inputs(node_type=JobCalculation)
            n_parents = len(parent_calcs)
            if n_parents != 1:
                raise UniquenessError(
                    "Input RemoteData is child of {} "
                    "calculation{}, while it should have a single parent"
                    "".format(n_parents, "" if n_parents == 0 else "s"))
            parent_calc = parent_calcs[0]
            has_parent = True

            # check that it is a valid parent
            #self._check_valid_parent(parent_calc)


            # if inpgen calc do
            # check if folder from db given, or get folder from rep.
            # Parent calc does not has to be on the same computer.

            if isinstance(parent_calc, FleurCalculation):
                new_comp = self.get_computer()
                old_comp = parent_calc.get_computer()
                if new_comp.uuid != old_comp.uuid:
                    #dont copy files, copy files localy
                    copy_remotely = False
                    #raise InputValidationError(
                    #    "FleurCalculation must be launched on the same computer"
                    #    " of the parent: {}".format(old_comp.get_name()))
            elif isinstance(parent_calc, FleurinputgenCalculation):
                fleurinpgen = True
                new_comp = self.get_computer()
                old_comp = parent_calc.get_computer()
                if new_comp.uuid != old_comp.uuid:
                    #dont copy files, copy files localy
                    copy_remotely = False
            else:
                raise InputValidationError(
                    "parent_calc, must be either an 'inpgen calculation' or"
                    " a 'fleur calculation'.")

        # check existence of settings (optional)
        settings = inputdict.pop(self.get_linkname('settings'), None)
        #print('settings: {}'.format(settings))
        if settings is None:
            settings_dict = {}
        else:
            if not isinstance(settings, ParameterData):
                raise InputValidationError("settings, if specified, must be of "
                                           "type ParameterData")
            else:
                settings_dict = settings.get_dict()
        #check for for allowed keys, ignor unknown keys but warn.
        for key in settings_dict.keys():
            if key not in self._settings_keys:
                #TODO warrning
                self.logger.info(u"settings dict key {} for Fleur calculation"
                                 "not reconized, only {} are allowed."
                                 "".format(key, self._settings_keys))
        # Here, there should be no other inputs
        if inputdict:
            raise InputValidationError(
                "The following input data nodes are "
                "unrecognized: {}".format(inputdict.keys()))

        #TODO: Detailed check of FleurinpData
        # if certain files are there in fleurinpData.
        # from where to copy

        ##############################
        # END OF INITIAL INPUT CHECK #


        # file copy stuff TODO check in fleur input (fleur modes tries to do this. but it is incomplete.
        # also what file fleur produces changes currently a lot...
        
        if has_fleurinp:
            self._DEFAULT_INPUT_FILE = fleurinp.get_file_abs_path(self._INPXML_FILE_NAME)

            #local_copy_list.append((
            #    fleurinp.get_file_abs_path(self._INPXML_FILE_NAME),
            #    self._INPXML_FILE_NAME))
            #copy ALL files from inp.xml
            allfiles = fleurinp.files
            for file1 in allfiles:
                local_copy_list.append((
                    fleurinp.get_file_abs_path(file1),
                    file1))
            modes = fleurinp.get_fleur_modes()

            # add files to mode_retrieved_filelist
            if modes['band']:
                mode_retrieved_filelist.append(self._BAND_FILE_NAME)
                mode_retrieved_filelist.append(self._BAND_GNU_FILE_NAME)
            if modes['dos']:
                mode_retrieved_filelist.append(self._DOS_FILE_NAME)
            if modes['forces']:
                print('FORCES!!!')
                mode_retrieved_filelist.append(self._NEW_XMlINP_FILE_NAME)
                mode_retrieved_filelist.append(self._FORCE_FILE_NAME)
            if modes['ldau']:
                mode_retrieved_filelist.append(self._NMMPMAT_FILE_NAME)
            #if noco, ldau, gw...
            # TODO: check from where it was copied, and copy files of its parent
            # if needed
        #self.logger.info("@@@@@@@@@@@@@@@@@@@@@@@@has_parent {}".format(has_parent))

        if has_parent:
            # copy the right files #TODO check first if file, exist and throw
            # warning, now this will throw an error
            outfolderpath = parent_calc.out.retrieved.folder.abspath
            self.logger.info(u"out folder path {}".format(outfolderpath))

            if fleurinpgen and (not has_fleurinp):
                for file1 in self._copy_filelist_inpgen:
                    local_copy_list.append((
                        os.path.join(outfolderpath, 'path', file1),
                        os.path.join(file1)))
            elif not fleurinpgen and (not has_fleurinp): # fleurCalc
                if with_hdf5:
                    copylist = self._copy_filelist_scf2_1
                else:
                    copylist = self._copy_filelist_scf
                for file1 in copylist:
                    local_copy_list.append((
                        os.path.join(outfolderpath, 'path', file1[0]),
                        os.path.join(file1[1])))
                filelist_tocopy_remote = filelist_tocopy_remote# + self._copy_filelist_scf_remote
                #TODO get inp.xml from parent fleurinpdata, since otherwise it will be doubled in repo
            elif fleurinpgen and has_fleurinp:
                # everything is taken care of
                pass
            elif not fleurinpgen and has_fleurinp:
                # input file is already taken care of
                if with_hdf5:
                    copylist = self._copy_filelist_scf2
                else:
                    copylist = self._copy_filelist_scf1
                for file1 in copylist:
                    local_copy_list.append((
                        os.path.join(outfolderpath, 'path', file1[0]),
                        os.path.join(file1[1])))
                filelist_tocopy_remote = filelist_tocopy_remote# + self._copy_filelist_scf_remote

            # TODO not on same computer -> copy needed files from repository,
            # if they are not there, throw error
            if copy_remotely: # on same computer.
                #print('copy files remotely')

                # from fleurmodes
                #if modes.get(['pot8']: #pot8 was removed from input
                #    filelist_tocopy_remote = filelist_tocopy_remote + self._copy_filelist_scf_remote
                #    filelist_tocopy_remote.append(self._POT_FILE_NAME)
                #    #filelist_tocopy_remote.append(self._POT2_FILE_NAME)
                if modes['dos']:
                    pass
                elif modes['band']:
                    pass
                else:
                    filelist_tocopy_remote = filelist_tocopy_remote + self._copy_filelist_scf_remote
                # from settings, user specified
                #TODO check if list?
                for file1 in settings_dict.get('additional_remotecopy_list', []):
                    filelist_tocopy_remote.append(file1)

                for file1 in settings_dict.get('remove_from_remotecopy_list', []):
                    if file1 in filelist_tocopy_remote:
                        filelist_tocopy_remote.remove(file1)

                for file1 in filelist_tocopy_remote:
                    remote_copy_list.append((
                        parent_calc_folder.get_computer().uuid,
                        os.path.join(parent_calc_folder.get_remote_path(), file1),
                        self._OUTPUT_FOLDER))

                #self.logger.info("remote copy file list {}".format(remote_copy_list))
        
        #  Prepare self._JUDFT_WARN_ONLY_INFO_FILE_NAME and upload
        # local_copy_list.append
        warn_only_filename = tempfolder.get_abs_path(self._JUDFT_WARN_ONLY_INFO_FILE_NAME)
        with open(warn_only_filename, 'w') as infile:
            infile.write("\n")


        ########## MAKE CALCINFO ###########

        calcinfo = CalcInfo()

        calcinfo.uuid = self.uuid
        # Empty command line by default
        #cmdline_params = settings_dict.pop('CMDLINE', [])
        #calcinfo.cmdline_params = (list(cmdline_params)
        #                           + ["-in", self._INPUT_FILE_NAME])

        self.logger.info(u"local copy file list {}".format(local_copy_list))

        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        #(remotemachinename, remoteabspath, relativedestpath)
        calcinfo.remote_symlink_list = remote_symlink_list
        #calcinfo.stdout_name = self._OUTPUT_FILE_NAME

        # Retrieve by default the output file and the xml file
        retrieve_list = []
        retrieve_list.append(self._OUTXML_FILE_NAME)
        retrieve_list.append(self._INPXML_FILE_NAME)
        #calcinfo.retrieve_list.append(self._OUTPUT_FILE_NAME)
        retrieve_list.append(self._SHELLOUTPUT_FILE_NAME)
        retrieve_list.append(self._ERROR_FILE_NAME)
        #calcinfo.retrieve_list.append(self._TIME_INFO_FILE_NAME)
        retrieve_list.append(self._OUT_FILE_NAME)
        if with_hdf5:
            retrieve_list.append(self._CDN_LAST_HDF5_FILE_NAME)
            retrieve_list.append(self._CDN1_FILE_NAME) # only for now because somthing is buggy
        else:
            retrieve_list.append(self._CDN1_FILE_NAME)
        #calcinfo.retrieve_list.append(self._INP_FILE_NAME)
        #calcinfo.retrieve_list.append(self._ENPARA_FILE_NAME)
        #calcinfo.retrieve_list.append(self._SYMOUT_FILE_NAME)
        #calcinfo.retrieve_list.append(self._KPTS_FILE_NAME)

        # if certain things are modefied, flags set,
        #other files should be retrieved, example DOS.x...

        for mode_file in mode_retrieved_filelist:
            retrieve_list.append(mode_file)
        #print('retrieve_list: {}'.format(retrieve_list))

        # user specific retrieve
        add_retrieve = settings_dict.get('additional_retrieve_list', [])
        #print('add_retrieve: {}'.format(add_retrieve))
        for file1 in add_retrieve:
            retrieve_list.append(file1)

        remove_retrieve = settings_dict.get('remove_from_retrieve_list', [])
        for file1 in remove_retrieve:
            if file1 in retrieve_list:
                retrieve_list.remove(file1)

        calcinfo.retrieve_list = []
        for file1 in retrieve_list:
            calcinfo.retrieve_list.append(file1)

        codeinfo = CodeInfo()
        # should look like: codepath -xmlInput < inp.xml > shell.out 2>&1
        walltime_sec = self.get_max_wallclock_seconds()
        #self.logger.info("!!!!!!!!!!!!!!!!!!! walltime_sec : {}"
        #                         "".format(walltime_sec))
        cmdline_params = []#, "-wtime", "{}".format(walltime_sec)]"-xml"
        #walltime_sec = self.get_max_wallclock_seconds()
        #print('walltime: {}'.format(walltime_sec))
        if with_hdf5:
            cmdline_params.append("-last_extra")

        if walltime_sec:
            walltime_min = max(1, walltime_sec/60)
            cmdline_params.append("-wtime")
            cmdline_params.append("{}".format(walltime_min))


        # user specific commandline_options
        for command in settings_dict.get('cmdline', []):
            cmdline_params.append(command)
        
        # Error tracking
        #cmdline_params.append("-g")
        #cmdline_params.append("-traceback")
        #cmdline_params.append("-debugtime")
        
        codeinfo.cmdline_params = list(cmdline_params)
        # + ["<", self._INPXML_FILE_NAME,
	    # ">", self._SHELLOUTPUT_FILE_NAME, "2>&1"]
        codeinfo.code_uuid = code.uuid
        codeinfo.withmpi = self.get_withmpi()
        codeinfo.stdin_name = None#self._INPUT_FILE_NAME
        codeinfo.stdout_name = self._SHELLOUTPUT_FILE_NAME
        #codeinfo.join_files = True
        codeinfo.stderr_name = self._ERROR_FILE_NAME

        calcinfo.codes_info = [codeinfo]
        '''
        # not needed in new version
        if fleurinpgen:# execute twice, as long start density stop
            codeinfo1 = CodeInfo()
            cmdline_params = ["-xmlInput"]
            codeinfo1.cmdline_params = list(cmdline_params)
            # + ["<", self._INPUT_FILE_NAME])#,
	        # ">",self._OUTPUT_FILE_NAME]
            codeinfo1.code_uuid = code.uuid
            codeinfo1.withmpi = self.get_withmpi()
            codeinfo1.stdin_name = self._INPUT_FILE_NAME
            codeinfo1.stdout_name = self._SHELLOUTPUT_FILE_NAME
            calcinfo.codes_info.append(codeinfo1)

        if settings_dict:
            raise InputValidationError("The following keys have been found in "
                "the settings input node, but were not understood: {}".format(
                ",".join(settings_dict.keys())))
        '''
        return calcinfo

    def _check_valid_parent(self, calc):
        """
        Check that calc is a valid parent for a FleurCalculation.
        It can be a FleurCalculation, InpgenCalculation, or (if the class exists) a
        CopyonlyCalculation
        :TODO: maybe assume that CopyonlyCalculation class always exists?
        """
        #from aiida.orm.calculation.job.simpleplugins.copyonly import CopyonlyCalculation

        try:
            if (((not isinstance(calc, FleurCalculation)))
                            and (not isinstance(calc, FleurinputgenCalculation))):
                            #and (not isinstance(calc, CopyonlyCalculation)) ):
                raise ValueError("Parent calculation must be a FleurCalculation, a "
                                 "FleurinputgenCalculation or a CopyonlyCalculation")
        except ImportError:
            if ((not isinstance(calc, FleurCalculation))
                            and (not isinstance(calc, FleurinputgenCalculation)) ):
                raise ValueError("Parent calculation must be a FleurCalculation or "
                                 "a FleurinputgenCalculation")


    def use_fleurinp(self, fleurinp):
        """
        set fleurinpdata and path to inp.xml file.
        """
        if not isinstance(fleurinp, FleurinpData):
            raise InputValidationError("The FleurinpData node given is not of type"
                                           " FleurinpData.")

        self._DEFAULT_INPUT_FILE = fleurinp.get_file_abs_path(self._INPXML_FILE_NAME)
        # somehow this is not working...
        self.use_fleurinpdata(fleurinp)


    def use_parent_calculation(self, calc):
        """
        Set the parent calculation of Fleur,
        from which it will inherit the outputsubfolder.
        The link will be created from parent RemoteData to FleurCalculation
        """
        from aiida.common.exceptions import NotExistent

        self._check_valid_parent(calc)

        remotedatas = calc.get_outputs(type=RemoteData)
        if not remotedatas:
            raise NotExistent("No output remotedata found in "
                                  "the parent")
        if len(remotedatas) != 1:
            raise UniquenessError("More than one output remotedata found in "
                                  "the parent")
        remotedata = remotedatas[0]

        self._set_parent_remotedata(remotedata)

    def _set_parent_remotedata(self,remotedata):
        """
        Used to set a parent remotefolder in the restart of fleur.
        """
        if not isinstance(remotedata,RemoteData):
            raise ValueError('remotedata must be a RemoteData')

        # complain if another remotedata is already found
        input_remote = self.get_inputs(node_type=RemoteData)
        if input_remote:
            raise ValidationError("Cannot set several parent calculation to a "
                                  "fleur calculation")

        self.use_parent_folder(remotedata)

    def _create_warnonly_file(self,calc):
        """
        Used to create a WARN_ONLY file in the calculation dir
        """
        pass



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
contains verdi commands for the scf workchain
in general these should become options of verdi aiida-fleur workchains 
"""

import click


@click.group()
def scf_wc():
    pass

@scf_wc.command()
def res():
    """
    Prints the result node to screen
    """
    click.echo('verdi aiida-fleur scf res')

@scf_wc.command()
def show():
    """
    plots the results of a 
    """
    click.echo('verdi aiida-fleur scf show')

@scf_wc.command()
def list():
    """
    similar to the verdi work list command, but this displays also some 
    specific information about the scfs
    """
    click.echo('verdi aiida-fleur scf list')


# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module is the Fleurinpmodefier class, which is used to manipulate
fleurinpdata objects in a way which keeps the proverance.
"""

# TODO implement undo
# TODO general clean up
# TODO discuss design again what is really needed and the default way to du things
import os
import re
from lxml import etree
from lxml.etree import XMLSyntaxError

from aiida.orm import DataFactory
#from aiida.workflows2.wf import wf
from aiida.work import workfunction as wf

FleurinpData = DataFactory('fleur.fleurinp')

class FleurinpModifier(object):

    def __init__(self, original):
        #print(type(original), type(FleurinpData))
        assert isinstance(original, FleurinpData), "Wrong AiiDA data type"

        self._original = original
        self._tasks = []

    @staticmethod
    @wf
    def modify_fleurinpdata(original, modifications):
        """
        WF, original must be a fleurinp data, modifications a fleurinp data as well
        modification a python dict of the form {'task':

        modifications parameter data of the form: {'tasks:
        out: a modified fleurinp data
        """

        # copy
        # get schema
        # read in inp.xml
        # add modifications
        # validate
        # save inp.xml
        # store new fleurinp (copy)

        new_fleurinp = original.copy()
        # TODO test if file is there!
        inpxmlfile = new_fleurinp.get_file_abs_path('inp.xml')
        modification_tasks = modifications.get_dict()['tasks']

        xmlschema_doc = etree.parse(new_fleurinp._schema_file_path)
        xmlschema = etree.XMLSchema(xmlschema_doc)
        parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)

        tree = etree.parse(inpxmlfile)
        # there is a bug when validating at parsetime, therefore we only
        #validate at parse time if file is invalid, to get nice error message
        if not xmlschema.validate(tree):
            tree = etree.parse(inpxmlfile, parser)

        new_fleurtree = FleurinpModifier.apply_modifications(fleurinp_tree_copy=tree,
            modification_tasks=modification_tasks)

        inpxmlfile = os.path.join(
                         new_fleurinp._get_folder_pathsubfolder.abspath, 'temp_inp.xml')
        new_fleurtree.write(inpxmlfile)

        new_fleurinp.del_file('inp.xml')
        new_fleurinp._add_path(str(inpxmlfile), 'inp.xml')
        os.remove(inpxmlfile)

        # default label and description
        new_fleurinp.label = 'mod_fleurinp'# {}'.format(original.label)
        new_fleurinp.description = 'Fleurinpdata with modifications (see inputs of modify_fleurinpdata)'
        #return {'modified_fleurinp' : new_fleurinp} # this will break other stuff (Scf), also link was not renamed somehow.
        return new_fleurinp

    @staticmethod
    def apply_modifications(fleurinp_tree_copy, modification_tasks, schema_tree=None):
        """
        lxml etree of inp.xml
        and task dictionary
        """
        from aiida_fleur.tools.xml_util import xml_set_attribv_occ,xml_set_first_attribv,xml_set_all_attribv, xml_set_text, xml_set_all_text, create_tag, replace_tag, delete_tag, delete_att, set_species, change_atomgr_att#, set_inpchanges
        from aiida_fleur.tools.xml_util import add_num_to_att

        def xml_set_attribv_occ1(fleurinp_tree_copy, xpathn, attributename, attribv, occ=[0], create=False):
            xml_set_attribv_occ(fleurinp_tree_copy, xpathn, attributename, attribv, occ=occ, create=create)
            return fleurinp_tree_copy

        def xml_set_first_attribv1(fleurinp_tree_copy, xpathn, attributename, attribv, create=False):
            xml_set_first_attribv(fleurinp_tree_copy, xpathn, attributename, attribv, create=create)
            return fleurinp_tree_copy

        def xml_set_all_attribv1(fleurinp_tree_copy, xpathn, attributename, attribv, create=False):
            xml_set_all_attribv(fleurinp_tree_copy, xpathn, attributename, attribv, create=create)
            return fleurinp_tree_copy

        def xml_set_text1(fleurinp_tree_copy, xpathn, text, create=False):
            xml_set_text(fleurinp_tree_copy, xpathn, text, create=create)
            return fleurinp_tree_copy

        def xml_set_all_text1(fleurinp_tree_copy, xpathn, text, create=False):
            xml_set_all_text(fleurinp_tree_copy, xpathn, text, create=create)
            return fleurinp_tree_copy

        def create_tag1(fleurinp_tree_copy, xpath, newelement, create=False):
            fleurinp_tree_copy = create_tag(fleurinp_tree_copy, xpath, newelement, create=create)
            return fleurinp_tree_copy

        def delete_att1(fleurinp_tree_copy, xpath, attrib):
            fleurinp_tree_copy = delete_att(fleurinp_tree_copy, xpath, attrib)
            return fleurinp_tree_copy

        def delete_tag1(fleurinp_tree_copy, xpath):
            fleurinp_tree_copy = delete_tag(fleurinp_tree_copy, xpath)
            return fleurinp_tree_copy

        def replace_tag1(fleurinp_tree_copy, xpath, newelement):
            fleurinp_tree_copy = replace_tag(fleurinp_tree_copy, xpath, newelement)
            return fleurinp_tree_copy

        def set_species1(fleurinp_tree_copy, species_name, attributedict, create=False):
            fleurinp_tree_copy = set_species(fleurinp_tree_copy, species_name, attributedict, create=create)
            return fleurinp_tree_copy

        def change_atomgr_att1(fleurinp_tree_copy, attributedict, position=None, species=None,create=False):
            fleurinp_tree_copy = change_atomgr_att(fleurinp_tree_copy, attributedict, position=position, species=species,create=create)
            return fleurinp_tree_copy


        def add_num_to_att1(fleurinp_tree_copy, xpathn, attributename, set_val, mode='abs', occ=[0], create=False):
            fleurinp_tree_copy = add_num_to_att(fleurinp_tree_copy, xpathn, attributename, set_val, mode=mode, occ=occ, create=create)
            return fleurinp_tree_copy

        '''
        def set_inpchanges1(fleurinp_tree_copy, change_dict):
            #fleurinp_tree_copy = set_inpchanges(fleurinp_tree_copy, change_dict)
            return fleurinp_tree_copy
        '''
        def set_inpchanges1(fleurinp_tree_copy, change_dict):
            """
            Does changes directly on the inp.xml file. Afterwards
            updates the inp.xml file representation and the current inp_userchanges
            dictionary with the keys provided in the 'change_dict' dictionary.

            :param change_dict: a python dictionary with the keys to substitute.
                                It works like dict.update(), adding new keys and
                                overwriting existing keys.
            """
            from aiida_fleur.tools.xml_util import write_new_fleur_xmlinp_file, get_inpxml_file_structure

            #TODO if we still want tracking that way, have to get fleurinp in argument
            '''
            if self.inp_userchanges is None:
                self._set_attr('inp_userchanges', {})

            # store change dict, to trac changes
            currentchangedict = self.inp_userchanges
            currentchangedict.update(change_dict)
            self._set_attr('inp_userchanges', currentchangedict)

            # load file, if it does not exsist error will be thrown in routine
            inpxmlfile = self.get_file_abs_path('inp.xml')

            if self._has_schema:
               #schema file for validation will be loaded later
               pass
            elif self._schema_file_path != None:
                print('Warning: The User set the XMLSchema file path manually, your'
                      'inp.xml will be evaluated! If it fails it is your own fault!')
            else:
                print('Warning: No XMLSchema file was provided, your inp.xml file '
                      'will not be evaluated and parsed! (I should never get here)')

            #read in tree
            tree = etree.parse(inpxmlfile)
            '''
            tree = fleurinp_tree_copy
            #apply changes to etree
            xmlinpstructure = get_inpxml_file_structure()
            new_tree = write_new_fleur_xmlinp_file(tree, change_dict, xmlinpstructure)

            return new_tree


        def set_nkpts1(fleurinp_tree_copy, count, gamma):#_orgi

            kpointlist_xpath = '/fleurInput/calculationSetup/bzIntegration/kPointList'
            #kpoint_xpath = '/fleurInput/calculationSetup/bzIntegration/kPoint*'

            tree = fleurinp_tree_copy
            new_kpo = etree.Element('kPointCount', count="{}".format(count), gamma="{}".format(gamma))
            new_tree = replace_tag(tree, kpointlist_xpath, new_kpo)

            return new_tree

        '''
        def set_species1(fleurinp_tree_copy, species_name, attributedict):
            #fleurinp_tree_copy = set_species(fleurinp_tree_copy , species_name, attributedict)
            print('in set_species')
            return fleurinp_tree_copy

        def change_atom1(fleurinp_tree_copy, attrib, value, position=None, species=None):
            #fleurinp_tree_copy = change_atom(fleurinp_tree_copy, attrib, value, position=None, species=None)

            print('in change_atom')
            return fleurinp_tree_copy

        def set_xpath1(fleurinp_tree_copy, xpath, value):
            #fleurinp_tree_copy = set_xpath(fleurinp_tree_copy, xpath, value)

            print('in set_xpath')
            return fleurinp_tree_copy
        '''

        actions = {
            'xml_set_attribv_occ' : xml_set_attribv_occ1,
            'xml_set_first_attribv' : xml_set_first_attribv1,
            'xml_set_all_attribv' : xml_set_all_attribv1,
            'xml_set_text' : xml_set_text1,
            'xml_set_all_text' : xml_set_all_text1,
            'create_tag' : create_tag1,
            'replace_tag' : replace_tag1,
            'delete_tag' : delete_tag1,
            'delete_att' : delete_att1,
            'set_species' : set_species1,
            'set_atomgr_att' : change_atomgr_att1,
            'set_inpchanges': set_inpchanges1,
            'set_nkpts' : set_nkpts1,
            'add_num_to_att' : add_num_to_att1

        }

        workingtree = fleurinp_tree_copy#.copy()
        if schema_tree:
            #xmlschema_doc = etree.parse(new_fleurinp._schema_file_path)
            xmlschema = etree.XMLSchema(schema_tree)
            parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)

        for task in modification_tasks:
            try:
                action = actions[task[0]]
            except KeyError:
                raise ValueError("Unknown task {}".format(task[0]))

            workingtree = action(workingtree, *task[1:])
            if schema_tree:
                if not xmlschema.validate(fleurinp_tree_copy):
                    pass# do something to get nice error message
                    # TODO maybe even delete wrong task
                    print('change not valid: {}'.format(task[1:]))
                else:
                    pass
        return workingtree

    def get_avail_actions(self):
        """
        returns the allowed functions from fleurinpmod, this is for checking purposes of other routines
        """
        outside_actions = {
            'xml_set_attribv_occ' : self.xml_set_attribv_occ,
            'xml_set_first_attribv' : self.xml_set_first_attribv,
            'xml_set_all_attribv' : self.xml_set_all_attribv,
            'xml_set_text' : self.xml_set_text,
            'xml_set_all_text' : self.xml_set_all_text,
            'create_tag' : self.create_tag,
            'replace_tag' : self.replace_tag,
            'delete_tag' : self.delete_tag,
            'delete_att' : self.delete_att,
            'set_species' : self.set_species,
            'set_atomgr_att' : self.set_atomgr_att,
            'set_inpchanges': self.set_inpchanges,
            'set_nkpts' : self.set_nkpts,
            'add_num_to_att' : self.add_num_to_att

        }
        return outside_actions

    '''
    def set_inpchanges(self, change_dict):
        self._tasks.append(('set_inpchanges', change_dict))

    def set_species(self, species_name, attributedict):
        self._tasks.append(('set_species', species_name, attributedict))

    def change_atom(self, attrib, value, position=None, species=None):
        self._tasks.append(('change_atom', attrib, value, position, species))

    def set_xpath(self, xpath, value):
        self._tasks.append(('set_xpath', xpath, value))
    '''
    def xml_set_attribv_occ(self, xpathn, attributename, attribv, occ=[0], create=False):
        self._tasks.append(('xml_set_attribv_occ', xpathn, attributename, attribv, occ, create))

    def xml_set_first_attribv(self, xpathn, attributename, attribv, create=False):
        self._tasks.append(('xml_set_first_attribv', xpathn, attributename, attribv, create))

    def xml_set_all_attribv(self, xpathn, attributename, attribv, create=False):
        self._tasks.append(('xml_set_all_attribv', xpathn, attributename, attribv, create))

    def xml_set_text(self, xpathn, text, create=False):
        self._tasks.append(('xml_set_text', xpathn, text, create))

    def xml_set_all_text(self, xpathn, text, create=False):
        self._tasks.append(('xml_set_all_text', xpathn, text, create))

    def create_tag(self, xpath, newelement, create=False):
        self._tasks.append(('create_tag', xpath, newelement, create))

    def delete_att(self, xpath, attrib):
        self._tasks.append(('delete_att', xpath, attrib))

    def delete_tag(self, xpath):
        self._tasks.append(('delete_tag', xpath))

    def replace_tag(self, xpath, newelement):
        self._tasks.append(('replace_tag', xpath, newelement))

    def set_species(self, species_name, attributedict, create=False):
        self._tasks.append(('set_species', species_name, attributedict, create))

    def set_atomgr_att(self, attributedict, position=None, species=None, create=False):
        self._tasks.append(('set_atomgr_att', attributedict, position, species, create))

    #for now
    def set_inpchanges(self, change_dict):
        self._tasks.append(('set_inpchanges', change_dict))

    def set_nkpts(self, count, gamma='F'):
        self._tasks.append(('set_nkpts', count, gamma))

    def add_num_to_att(self, xpathn, attributename, set_val, mode='abs', occ=[0], create=False):
        self._tasks.append(('add_num_to_att', xpathn, attributename, set_val, mode, occ, create))
    #def set_attribute(self, key, value):
    #    pass

    #def set_inpchanges(self, attributedict):
    #    for k, v in attrib

    def validate(self):
        #print('in validate')
        inpxmlfile = self._original.get_file_abs_path('inp.xml')
        tree = etree.parse(inpxmlfile)

        try:# could be not found or on another computer...
            xmlschema_tree = etree.parse(self._original._schema_file_path)
            with_schema = True
        except:
            with_schema = False
            print('No schema file found')
            return
        if with_schema:
            tree = self.apply_modifications(tree, self._tasks, schema_tree=xmlschema_tree)
        return tree

    def show(self, display=True, validate=False):
        # apply modification


        if validate:
            tree = self.validate()
        else:
            inpxmlfile = self._original.get_file_abs_path('inp.xml')
            tree = etree.parse(inpxmlfile)
            tree = self.apply_modifications(tree, self._tasks)

        if display:
            xmltreestring = etree.tostring(tree, xml_declaration=True, pretty_print=True)
            print(xmltreestring)
        return tree
        self.logger.debug(self.apply_modifications(self._original.get_dict(), self._tasks))

    def changes(self):
        from pprint import pprint
        pprint(self._tasks)
        return self._tasks

    def freeze(self):
        modifications = DataFactory("parameter")(dict={"tasks": self._tasks})
        modifications.description = u'Fleurinpmodifier Tasks and inputs of these.'
        modifications.label = u'Fleurinpdata modifications'
        # This runs in a inline calculation to keep provenance
        out = self.modify_fleurinpdata(
            original=self._original,
            modifications=modifications,
            _label='fleurinp modifier',
            _description='This workfunction modified an Fleurinpdataobject')
        return out

    def undo(self, all=False):
        if all:
            self._tasks = []
        else:
            if self._tasks:
                self._tasks.pop()
                #del self._tasks[-1]
        return self._tasks

if __name__ == "__main__":

    P = FleurinpData
    '''
    orig = P(dict={'a': 1, 'b': True, 'c': "aaa"}).store()


    print("Original: {}".format(orig.pk))
    #print(json.dumps(orig.get_dict(), indent=2))


    modifier = fleurinpModifier(orig)


    new = modifier.freeze()
    print("new: {}".format(new.pk))
    #print(json.dumps(new.get_dict(), indent=2))
    '''








# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In here we put all things (methods) that are common to workflows AND
depend on AiiDA classes, therefore can only be used if the dbenv is loaded.
Util that does not depend on AiiDA classes should go somewhere else.
"""
from string import digits
from aiida.orm import DataFactory, Node, load_node, CalculationFactory

KpointsData =  DataFactory('array.kpoints')
RemoteData = DataFactory('remote')
ParameterData = DataFactory('parameter')
#FleurInpData = DataFactory('fleurinp.fleurinp')
FleurInpData = DataFactory('fleur.fleurinp')
FleurProcess = CalculationFactory('fleur.fleur').process()
FleurinpProcess = CalculationFactory('fleur.inpgen').process()


def is_code(code):
    """
    Test if the given input is a Code node, by object, id, uuid, or pk
    if yes returns a Code node in all cases
    if no returns None
    """
    from aiida.orm import Code, load_node
    #Test if Code
    if isinstance(code, Code):
        return code
    #Test if pk, if yes, is the corresponding node Code
    pk = None
    try:
        pk=int(code)
    except:
        pass
    if pk:
        code = load_node(pk)
        if isinstance(code, Code):
            return code
        else:
            return None
    #given as string
    codestring = None
    try:
        codestring = str(code)
    except:
        pass
    if codestring:
        code = Code.get_from_string(codestring)
        return code
    #Test if uuid, if yes, is the corresponding node Code
    # TODO: test for uuids not for string (guess is ok for now)
    '''
    uuid = None
    try:
        uuid = str(code)
    except:
        pass
    if uuid:
        code = load_node(uuid)
        if isinstance(code, Code):
            return code
        else:
            return None
    '''
    return None


def get_inputs_fleur(code, remote, fleurinp, options, label='', description='', settings=None, serial=False):
    """
    get the input for a FLEUR calc
    """
    inputs = FleurProcess.get_inputs_template()
    #print('Template fleur {} '.format(inputs))
    if remote:
        inputs.parent_folder = remote
    if code:
        inputs.code = code
    if fleurinp:
        inputs.fleurinpdata = fleurinp

    for key, val in options.iteritems():
        if val==None:
            continue
        else:
            inputs._options[key] = val

    if description:
        inputs['_description'] = description
    else:
        inputs['_description'] = ''
    if label:
        inputs['_label'] = label
    else:
        inputs['_label'] = ''
    #TODO check  if code is parallel version?
    if serial:
        inputs._options.withmpi = False # for now
        inputs._options.resources = {"num_machines": 1}

    if settings:
        inputs.settings = settings
    
    return inputs


def get_inputs_inpgen(structure, inpgencode, options, label='', description='', params=None):
    """
    get the input for a inpgen calc
    """
    inputs = FleurinpProcess.get_inputs_template()
    #print('Template inpgen {} '.format(inputs))

    if structure:
        inputs.structure = structure
    if inpgencode:
        inputs.code = inpgencode
    if params:
        inputs.parameters = params
    for key, val in options.iteritems():
        if val==None:
            #leave them out, otherwise the dict schema won't validate
            continue
        else:
            inputs._options[key] = val

    if description:
        inputs['_description'] = description
    else:
        inputs['_description'] = ''

    if label:
        inputs['_label'] = label
    else:
        inputs['_label'] = ''

    #inpgen run always serial
    inputs._options.withmpi = False # for now
    inputs._options.resources = {"num_machines": 1}
    #print(inputs)
    return inputs

'''
def get_inputs_fleur(code, remote, fleurinp, options, label='', description='', settings=None, serial=False, **kwargs):
    """
    get the input for a FLEUR calc
    """
    inputs = FleurProcess.get_inputs_template()
    #print('Template fleur {} '.format(inputs))
    if remote:
        inputs.parent_folder = remote
    if code:
        inputs.code = code
    if fleurinp:
        inputs.fleurinpdata = fleurinp

    #for key, val in options.iteritems():
    #    if val==None:
    #        continue
    #    elif isinstance(val, str):# ensure unicode 
    #        inputs.options[key] = unicode(val)
    #    else:
    #        inputs.options[key] = val

    if description:
        inputs._description = description
    else:
        inputs._description = ''
    if label:
        inputs._label = label
    else:
        inputs._label = ''
    #TODO check  if code is parallel version?
    if serial:
        if not options:
            options = {}
        options['withmpi'] = False # for now
        #TODO not every machine/scheduler type takes number of machines
        #  lsf takes number of total_mpi_procs,slurm and psb take num_machinces,\
        # also a full will run here mpi on that node... also not what we want.ß
        options['resources'] = {"num_machines": 1}

    if settings:
        inputs.settings = settings

    if options:
        inputs._options = options
    
    # Currently this does not work, find out howto...
    #for key, val in kwargs.iteritems(): 
    #    inputs[key] = val
    """
    options = {
    "max_wallclock_seconds": int,
    "resources": dict,
    "custom_scheduler_commands": unicode,
    "queue_name": basestring,
    "computer": Computer,
    "withmpi": bool,
    "mpirun_extra_params": Any(list, tuple),
    "import_sys_environment": bool,
    "environment_variables": dict,
    "priority": unicode,
    "max_memory_kb": int,
    "prepend_text": unicode,
    "append_text": unicode}
    """
    return inputs


def get_inputs_inpgen(structure, inpgencode, options, label='', description='', params=None, **kwargs):
    """
    get the input for a inpgen calc
    """
    inputs = FleurinpProcess.get_inputs_template()
    #print('Template inpgen {} '.format(inputs))

    if structure:
        inputs.structure = structure
    if inpgencode:
        inputs.code = inpgencode
    if params:
        inputs.parameters = params
    #for key, val in options.iteritems():
    #    if val==None:
    #        #leave them out, otherwise the dict schema won't validate
    #        continue
    #    else:
    #        inputs.options[key] = val

        
    if description:
        inputs._description = description
    else:
        inputs._description = ''

    if label:
        inputs._label = label
    else:
        inputs._label = ''

    if not options:
        options = {}
    #inpgen run always serial
    options[u'withmpi'] = False # for now
    options[u'resources'] = {u"num_machines": 1}
    #print(inputs)
    if options:
        inputs._options = options
    
    # Currently this does not work, find out howto...
    #for key, val in kwargs.items():
    #    inputs[key] = val
    
    return inputs
'''



def choose_resources_fleur(nkpt, natm, max_resources={"num_machines": 1}, ncores_per_node=24, memmory_gb=120):
    """
    nkpt: int, number of kpoints
    natm: int, number of atoms in the cell (for basis estimation, i.e number (valence) of electrons)
    max_resources: dict, maximum computing resource to choose from
    ncores_per_node, int how many cores are there per node
    memmory_gb: how much memmory in GB is there on one node?
    
    returns nodes, mpi_per_node, openmp, warnings: int, int, int, list
 
    # TODO has to be refinded for > 1 node systems (larger) systems and memmory requirements
    # often to many nodes are currently chooses for a medium system
    """
    # if claix, ram and cpus:
    # if jureca 
    # if jureca booster
    # 
    from aiida_fleur.tools.decide_ncore import gcd
    
    ncores_per_node = ncores_per_node
    memmory_gb = memmory_gb
    warnings = []
    
    # for fleur
    if natm > 1000:
        nodes = 64*nkpt # total nodes
        mpi_per_node = 2
        openmp = ncores_per_node/mpi_per_node
    elif natm > 500:
        nodes = 16*nkpt # total nodes
        mpi_per_node = 2
        openmp = ncores_per_node/mpi_per_node
    elif natm > 200:
        nodes = 4*nkpt # total nodes
        mpi_per_node = 4
        openmp = ncores_per_node/mpi_per_node
    elif natm > 80:
        if nkpt < 10:
            nodes = 2*nkpt # total nodes
            mpi_per_node = 4
            openmp = ncores_per_node/mpi_per_node
        elif nkpt < 100:
            nodes = nkpt # total nodes
            mpi_per_node = 4
            openmp = ncores_per_node/mpi_per_node        
        else:
            factor = 1
            nodes = nkpt/factor # total nodes
            mpi_per_node = 4
            openmp = ncores_per_node/mpi_per_node
    elif natm < 30:
        factor = gcd(ncores_per_node, nkpt)
        nodes = 1
        mpi_per_node = factor
        openmp = ncores_per_node/mpi_per_node          
    else:
        if nkpt < 20:
            factor = gcd(ncores_per_node, nkpt)
            nodes = 1
            mpi_per_node = factor
            openmp = ncores_per_node/mpi_per_node
        else:
            factor = gcd(ncores_per_node, nkpt)
            nodes = nkpt/factor
            mpi_per_node = factor
            openmp = ncores_per_node/mpi_per_node  
            
    #print nodes
    #print mpi_per_node
    #print openmp
    
    if max_resources:
        max_numnodes = max_resources.get('num_machines', None)
        max_mpiproc = max_resources.get("tot_num_mpiprocs", None)
        if (max_numnodes is not None) and (max_mpiproc is None):
            if max_numnodes < nodes:
                warning = ('The max number of provided compute nodes {} is less then the recommened {}'
                      ', consider providing more resouces for this calculation'
                      'or using less kpoints.'.format(max_numnodes, nodes))
                nodes = max_numnodes
                warnings.append(warning)
                print(warning)
        elif max_mpiproc is not None:
            mpiproc = mpi_per_node*nodes
            if mpiproc >= max_mpiproc:
                nodes = max(max_mpiproc/mpi_per_node,1) # should be at least 1...
            # else everything is fine
        else:
            max_number_total_mpi_proc =  max_resources.get('num_machines', None)
            
    return nodes, mpi_per_node, openmp, warnings

def get_scheduler_extras(code, resources, extras={}, project='jara0172'):
    """
    This is a utilty function with the goal to make prepare the right resource and scheduler extras for a given computer.
    Since this is user dependend you might want to create your own.

    return: dict, custom scheduler commands
    """
    nnodes = resources.get('num_machines', 1)
    
    # TODO memmory has to be done better...
    mem_per_node = 120000# max recommend 126000 MB on claix jara-clx nodes
    mem_per_process = mem_per_node/24
    if not extras:
        # use defaults # TODO add other things, span, pinnning... openmp
        extras = {'lsf' : '#BSUB -P {} \n#BSUB -M {}  \n#BSUB -a intelmpi'.format(project, mem_per_process),#{'-P' : 'jara0043', '-M' : memp_per_node*nnodes, '-a' : 'intelmpi'},
                 'torque' : '',#{},
                 'direct' : ''}#{}}

    # get the scheduler type from the computer the code is run on.
    com = code.get_computer()
    #com_name = com.get_name()
    scheduler_type = com.get_scheduler_type()

    default_per_machine = com.get_default_mpiprocs_per_machine()
    if not default_per_machine:
        default_per_machine = 24# claix, lsf does can not have default mpiprocs... #TODO this better
    tot_num_mpiprocs = resources.get('tot_num_mpiprocs', default_per_machine*nnodes)

    if scheduler_type == 'lsf':
        new_resources = {'tot_num_mpiprocs' : tot_num_mpiprocs}# only this needs to be given
    elif scheduler_type == 'torque':
        new_resources = resources#{'num_machines', 1} # on iff003 currently we do not do multinode mpi,
        #like this it will get stuck on iff003
    else:
        new_resources = resources
    scheduler_extras = extras.get(scheduler_type, '')

    return new_resources, scheduler_extras


#test
###############################
#codename = 'inpgen@local_mac'#'inpgen_v0.28@iff003'#'inpgen_iff@local_iff'
#codename2 = 'fleur_v0.28@iff003'#'fleur_mpi_v0.28@iff003'# 'fleur_iff_0.28@local_iff''
#codename2 = 'fleur_max_1.3_dev@iff003'
#codename2 = 'fleur_mpi_max_1.3_dev@iff003'
#codename4 = 'fleur_mpi_v0.28@claix'
###############################
#code = Code.get_from_string(codename)
#code2 = Code.get_from_string(codename2)
#code4 = Code.get_from_string(codename4)
#print(get_scheduler_extras(code, {'num_machines' : 1}))
#print(get_scheduler_extras(code2, {'num_machines' : 2}))
#print(get_scheduler_extras(code4, {'num_machines' : 1}))

def test_and_get_codenode(codenode, expected_code_type, use_exceptions=False):
    """
    Pass a code node and an expected code (plugin) type. Check that the
    code exists, is unique, and return the Code object.

    :param codenode: the name of the code to load (in the form label@machine)
    :param expected_code_type: a string with the plugin that is expected to
      be loaded. In case no plugins exist with the given name, show all existing
      plugins of that type
    :param use_exceptions: if True, raise a ValueError exception instead of
      calling sys.exit(1)
    :return: a Code object
    """
    import sys
    from aiida.common.exceptions import NotExistent
    from aiida.orm import Code


    try:
        if codenode is None:
            raise ValueError
        code = codenode
        if code.get_input_plugin_name() != expected_code_type:
            raise ValueError
    except (NotExistent, ValueError):
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(Code,
                  filters={'attributes.input_plugin':
                               {'==': expected_code_type}},
                  project='*')

        valid_code_labels = ["{}@{}".format(c.label, c.get_computer().name)
                             for [c] in qb.all()]

        if valid_code_labels:
            msg = ("Pass as further parameter a valid code label.\n"
                   "Valid labels with a {} executable are:\n".format(
                expected_code_type))
            msg += "\n".join("* {}".format(l) for l in valid_code_labels)

            if use_exceptions:
                raise ValueError(msg)
            else:
                print(msg)#, file=sys.stderr)
                sys.exit(1)
        else:
            msg = ("Code not valid, and no valid codes for {}.\n"
                   "Configure at least one first using\n"
                   "    verdi code setup".format(
                expected_code_type))
            if use_exceptions:
                raise ValueError(msg)
            else:
                print(msg)#, file=sys.stderr)
                sys.exit(1)

    return code


def get_kpoints_mesh_from_kdensity(structure, kpoint_density):
    """
    params: structuredata, Aiida structuredata
    params: kpoint_density

    returns: tuple (mesh, offset)
    returns: kpointsdata node
    """
    kp = KpointsData()
    kp.set_cell_from_structure(structure)
    density  = kpoint_density #1/A
    kp.set_kpoints_mesh_from_density(density)
    mesh = kp.get_kpoints_mesh()
    return mesh, kp

# test
# print(get_kpoints_mesh_from_kdensity(load_node(structure(120)), 0.1))
#(([33, 33, 18], [0.0, 0.0, 0.0]), <KpointsData: uuid: cee9d05f-b31a-44d7-aa72-30a406712fba (unstored)>)
# mesh, kp = get_kpoints_mesh_from_kdensity(structuredata, 0.1)
#print mesh[0]

# TODO maybe allow lists of uuids in workchain dict, or write a second funtion for this,...
# The question is how do get the 'enthalpy for a reaction out of my database?
# where I have redundant calculations or calculations with different parameters...
# are total energies comparable?
# -> as long as the same scheme ist used (all GGA or all GGA+U)
# total energies are compareable and the gibs enthalpy is approximatily the
# total energy difference
# there are tricks to also compare mixed energies, with experimental fits
# for binary reactions, where both is needed

def determine_favorable_reaction(reaction_list, workchain_dict):
    """
    Finds out with reaction is more favorable by simple energy standpoints

    # TODO check physics
    reaction list: list of reaction strings
    workchain_dict = {'Be12W' : uuid_wc or output, 'Be2W' : uuid, ...}

    return dictionary that ranks the reactions after their enthalpy
    """
    from aiida.orm import load_node
    from aiida.orm.calculation.work import WorkCalculation
    from aiida_fleur.tools.common_fleur_wf_util import get_enhalpy_of_equation
    # for each reaction get the total energy sum
    # make sure to use the right multipliers...
    # then sort the given list from (lowest if negativ energies to highest)
    energy_sorted_reactions = []
    formenergy_dict ={}
    for compound, uuid in workchain_dict.iteritems():
        # TODO ggf get formation energy from ouput node, or extras
        if isinstance(uuid, float):# allow to give values
            formenergy_dict[compound] = uuid
            continue
        n = load_node(uuid)
        extras = n.get_extras() # sadly there is no get(,) method...
        try:
            formenergy = extras.get('formation_energy', None)
        except KeyError:
            formenergy = None
        if not formenergy: # test if 0 case ok
            if isinstance(n, WorkCalculation):
                plabel = n.get_attr('_process_label')
                if plabel == 'fleur_initial_cls_wc':
                    try:
                        ouputnode = n.out.output_initial_cls_wc_para.get_dict()
                    except AttributeError:
                        try:
                            ouputnode = n.out.output_inital_cls_wc_para.get_dict()
                        except:
                            ouputnode = None
                            formenergy = None
                            print('WARNING: ouput node of {} not found. I skip'.format(n))
                            continue
                    formenergy = ouputnode.get('formation_energy')
                    # TODO is this value per atom?
                else: # check if corehole wc?
                     pass

        formenergy_dict[compound] = formenergy


    for reaction_string in reaction_list:
        ent_peratom = get_enhalpy_of_equation(reaction_string, formenergy_dict)
        print(ent_peratom)
        energy_sorted_reactions.append([reaction_string, ent_peratom])
    energy_sorted_reactions = sorted(energy_sorted_reactions, key=lambda ent: ent[1])
    return energy_sorted_reactions


# test
#reaction_list = ['1*Be12W->1*Be12W', '2*Be12W->1*Be2W+1*Be22W', '11*Be12W->5*W+6*Be22W', '1*Be12W->12*Be+1*W', '1*Be12W->1*Be2W+10*Be']
#workchain_dict = {'Be12W' : '4f685bc5-b5fb-46d3-aad6-e0f512c3313d',
#                  'Be2W' : '045d3071-f442-46b4-8d6b-3c85d72b24d4',
#                  'Be22W' : '1e32880a-bdc9-4081-a5da-be04860aa1bc',
#                  'W' : 'f8b12b23-0b71-45a1-9040-b51ccf379439',
#                  'Be' : 0.0}
#reac_list = determine_favorable_reaction(reaction_list, workchain_dict)
#print reac_list
#{'products': {'Be12W': 1}, 'educts': {'Be12W': 1}}
#0.0
#{'products': {'Be2W': 1, 'Be22W': 1}, 'educts': {'Be12W': 2}}
#0.114321037514
#{'products': {'Be22W': 6, 'W': 5}, 'educts': {'Be12W': 11}}
#-0.868053153884
#{'products': {'Be': 12, 'W': 1}, 'educts': {'Be12W': 1}}
#-0.0946046496213
#{'products': {'Be': 10, 'Be2W': 1}, 'educts': {'Be12W': 1}}
#0.180159355144
#[['11*Be12W->5*W+6*Be22W', -0.8680531538839534], ['1*Be12W->12*Be+1*W', -0.0946046496213127], ['1*Be12W->1*Be12W', 0.0], ['2*Be12W->1*Be2W+1*Be22W', 0.11432103751404535], ['1*Be12W->1*Be2W+10*Be', 0.1801593551436103]]


def performance_extract_calcs(calcs):
    """
    Extracts some runtime and system data from given fleur calculations
    
    :params calcs: list of calculation nodes/pks/or uuids. Fleur calc specific
    
    :returns data_dict: dictionary, dictionary of arrays with the same lengt, 
                        from with a panda frame can be created.
    
    Note: Is not the fastest for many calculations > 1000.
    """
    data_dict = {u'n_symmetries':[], u'n_spin_components' : [],
                 u'n_kpoints': [], u'n_iterations': [], 
                 u'walltime_sec' : [], u'walltime_sec_per_it' : [], 
                 u'n_iterations_total' : [], 
                 u'density_distance': [], u'computer':[],
                 u'n_atoms' : [], u'kmax':[],
                 u'cost' : [], u'costkonstant' : [], 
                 u'walltime_sec_cor' : [], u'total_cost' : [],
                 u'fermi_energy' : [], u'bandgap' : [], 
                 u'energy' : [], u'force_largest' : [], 
                 u'ncores' : [], u'pk' : [],
                 u'uuid' : [], u'serial' : [],
                 u'resources' : []}
    count = 0
    for calc in calcs:
        if not isinstance(calc, Node):
            calc = load_node(calc)
        count = count + 1
        pk = calc.pk
        print(count, pk)
        res = calc.res
        res_keys = list(res)
        try:
            efermi = res.fermi_energy
        except AttributeError:
            print('skipping {}, {}'.format(pk, calc.uuid))
            continue # we skip these entries
            efermi = -10000

        try:
            gap = res.bandgap
        except AttributeError:
            gap = -10000   
            continue
            print('skipping 2 {}, {}'.format(pk, calc.uuid))


        try:
            energy = res.energy
        except AttributeError:
            energy = 0.0
            print('skipping 3 {}, {}'.format(pk, calc.uuid))
            continue

        data_dict['bandgap'].append(gap)        
        data_dict['fermi_energy'].append(efermi)
        data_dict['energy'].append(energy)
        data_dict['force_largest'].append(res.force_largest)
        data_dict['pk'].append(pk)
        data_dict['uuid'].append(calc.uuid)
        data_dict['n_symmetries'].append(res.number_of_symmetries)
        nspins = res.number_of_spin_components
        data_dict['n_spin_components'].append(nspins)
        nkpt = res.number_of_kpoints
        data_dict['n_kpoints'].append(nkpt)
        niter = res.number_of_iterations
        data_dict['n_iterations'].append(niter)
        data_dict['n_iterations_total'].append(res.number_of_iterations_total)



        if u'charge_density' in res_keys:
            data_dict['density_distance'].append(res.charge_density)
        else: # magnetic, old
            data_dict['density_distance'].append(res.overall_charge_density)


        walltime = res.walltime
        if walltime <= 0:
            # date was not considert yet, we assume one day...
            walltime_new = walltime + 86400
        else:
            walltime_new = walltime

        walltime_periteration = walltime_new/niter


        data_dict['walltime_sec'].append(walltime)
        data_dict['walltime_sec_cor'].append(walltime_new)
        data_dict['walltime_sec_per_it'].append(walltime_periteration)
        cname = calc.get_computer().name
        data_dict['computer'].append(cname)
        natom = res.number_of_atoms
        data_dict['n_atoms'].append(natom)

        fleurinp = calc.get_inputs_dict()['fleurinpdata']
        kmax = fleurinp.inp_dict['calculationSetup']['cutoffs']['Kmax']
        data_dict['kmax'].append(kmax)


        cost = calc_time_cost_function(natom, nkpt, kmax, nspins)
        total_cost = cost * niter

        serial = not calc.get_withmpi()
        #codename = calc.get_code().label
        #code_col.append(codename)
        
        #if 'mpi' in codename:
        #    serial = False
        #else:
        #    serial = True
        data_dict['serial'].append(serial)
        
        resources = calc.get_resources()
        mpi_proc = get_mpi_proc(resources)    

        c_ratio = cost_ratio(cost, walltime_new, mpi_proc)
        data_dict['resources'].append(resources)
        data_dict['cost'].append(cost)
        data_dict['costkonstant'].append(c_ratio)
        data_dict['total_cost'].append(total_cost)
        data_dict['ncores'].append(mpi_proc)  

    return data_dict


def get_mpi_proc(resources):
    nmachines = resources.get('num_machines', 0)
    total_proc = resources.get('tot_num_mpiprocs', 0)
    if not total_proc:
        if nmachines:
            total_proc = nmachines*resources.get('default_mpiprocs_per_machine', 12)
        else:
            total_proc = resources.get('tot_num_mpiprocs', 24)
        
    return total_proc

def calc_time_cost_function(natom, nkpt, kmax, nspins=1):
    costs = natom**3 * kmax**3 * nkpt * nspins
    return costs

def calc_time_cost_function_total(natom, nkpt, kmax, niter, nspins=1):
    costs = natom**3 * kmax**3 * nkpt * nspins * niter
    return costs

def cost_ratio(total_costs, walltime_sec, ncores):
    ratio = total_costs/(walltime_sec*ncores)
    return ratio
    
    
    
def get_paranode(struc, para_nodes):
    """
    find out if a parameter node for a structure is in para_nodes
    (currently very creedy, but lists are small (100x100) but maybe reduce database accesses)
    """

    suuid = struc.uuid
    formula = struc.get_formula()
    element = formula.translate(None, digits)
    #print para_nodes
    for para in para_nodes:
        struc_uuid = para.get_extra('struc_uuid', None)
        para_form = para.get_extra('formula', None)
        para_ele = para.get_extra('element', None)
        if suuid == struc_uuid:
            return para
        elif formula == para_form:
            return para
        elif element == para_ele:
            return para
        elif element == para_form:
            return para
        else:
            pass
            #Do something else (test if parameters for a certain element are there)
            #....
    # we found no parameter node for the given structure therefore return none
    return None
    



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
A dummy workchain to test nested workchains
"""
import time
from aiida.orm import Code, DataFactory
from aiida.work.workchain import WorkChain
from aiida.work.run import submit
from aiida.work.workchain import ToContext
from aiida.work.process_registry import ProcessRegistry
from aiida.work.workchain import Outputs
from aiida.orm.data.base import Str
RemoteData = DataFactory('remote')
StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')


class dummy_wc(WorkChain):

    def __init__(self, *args, **kwargs):
        super(dummy_wc, self).__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec):
        super(dummy_wc, cls).define(spec)
        spec.input("str_display", valid_type=Str, required=True),
        spec.outline(
            cls.display,
            cls.run_sub1,
            #cls.run_sub2,
            #cls.run_sub3,
            cls.return_out
        )
        spec.dynamic_output()

    def display(self):
        message = 'message from dummy_wc: {}'.format(self.inputs.str_display)
        self.report(message)

        self.ctx.sub1 = None
        self.ctx.sub2 = None
        self.ctx.sub31 = None
        self.ctx.sub32 = None
        time.sleep(65)


    def get_input(self):
        message = 'I am here'
        self.report(message)
        time.sleep(10)


    def run_sub1(self):
        """
        Submiting this subwork chain is still buggy somehow...
        """
        print('run_sub1')
        n = 3
        allres = {}
        time.sleep(35)
        for i in range(0,n):
            inputs_sub = Str('This is wonderful 1.{}'.format(i))
            res = submit(sub_dummy_wc, str_display=inputs_sub)
            label = 'sub1_run{}'.format(i)
            allres[label] = res
            time.sleep(65)
            self.get_input()
        return ToContext(**allres)


    def run_sub2(self):
        print('run_sub2')
        inputs_sub = Str('This is wonderful 2')
        res = sub_dummy_wc.run(str_display=inputs_sub)

        self.ctx.sub2 = res


    def return_out(self):
        message = 'generating output nodes'
        self.report(message)
        subout = '{} | {} | {}, {}'.format(self.ctx.sub1, self.ctx.sub2, self.ctx.sub31, self.ctx.sub32)
        self.report(subout)
        outdict = {'out_dummy_wc' : Str('wonderful dummy_wc')}
        for link_name, node in outdict.iteritems():
            self.out(link_name, node)


class sub_dummy_wc(WorkChain):

    def __init__(self, *args, **kwargs):
        super(sub_dummy_wc, self).__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec):
        super(sub_dummy_wc, cls).define(spec)
        spec.input("str_display", valid_type=Str, required=True),
        spec.outline(
            cls.display,
            cls.run_sub,
            cls.return_out
        )
        spec.dynamic_output()

    def display(self):
        message = 'message from sub_dummy_wc: {}'.format(self.inputs.str_display)
        self.report(message)
        time.sleep(65)
    def run_sub(self):
        pass
        # ggf to test further depth call this wc rekursively


    def return_out(self):
        time.sleep(65)
        message = 'generating output nodes sub_dummy'
        self.report(message)
        outdict = {'out_sub_dummy_wc' : Str('wonderful sub_dummy_wc')}
        for link_name, node in outdict.iteritems():
            self.out(link_name, node)


class dummy_wf():
    pass



import pytest


# is_code
def test_is_code_interface():
    from aiida.orm import Code, load_node
    from aiida_fleur.tools.common_fleur_wf import is_code
    from aiida.common.exceptions import NotExistent
    
    with pytest.raises(NotExistent):
        is_code(Code)
    
    #inpgen has to be stored in db
    # maybe do a verdi code setup
    #code_uuid = load_node(1).uuid
    #code_pk = 1
    #code_name = 'inpgen'
    
    #assert is_code(code_uuid)
    #assert is_code(code_pk)
    #assert is_code(code_name)


'''uncommented these 2 tests because the process builder input None is not currently not == None...
# get_inputs_fleur
def test_get_inputs_fleur_interface():
    from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur
    from aiida.orm import Code, DataFactory


    code = Code()#'code'
    fleurinp = DataFactory('fleur.fleurinp')()
    remote = DataFactory('remote')()
    options = {}
    label = 'test'
    description = 'test des'
    inputs = {'code' : code, 'fleurinp' :  fleurinp, 'remote' : remote, 'options' : options, 'label' : label, 'description' : description}
    returns = {'description' : 'test des', 'label' : 'test', 'options' : None, 'store_provenance' : True,
                'code' : code, 'fleurinpdata' : fleurinp, 'parent_folder' : remote, 'settings' : None}
    assert get_inputs_fleur(**inputs) == returns


# get_inputs_inpgen
def test_get_inputs_inpgen_interface():
    from aiida_fleur.tools.common_fleur_wf import get_inputs_inpgen
    from aiida.orm import Code, DataFactory

    structure = DataFactory('structure')()#'structure'
    code = Code()#'inpgencode'
    options = {}
    label = 'test'
    description = 'test des'
    inputs = {'structure' : structure, 'inpgencode' : code, 'options' : options, 'label' : label, 'description' : description}
    returns = {'description' : 'test des', 'label' : 'test', 'options' : {'withmpi': False, 'resources': {'num_machines': 1}},
               'store_provenance' : True, 'code' : code, 'parameters' : None, 'settings' : None, 'structure' : structure}
    assert get_inputs_inpgen(**inputs) == returns


'''
# get_scheduler_extras

# test_and_get_codenode
def test_test_and_get_codenode_inpgen():
    from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
    from aiida.orm import Code
    from aiida.common.exceptions import NotExistent

    # install code setup code
    code = Code(input_plugin_name='fleur.inpgen')
    code.label = 'inpgen'
    #code = Code.get_from_string('inpgen@localhost')
    expected = 'fleur.inpgen'
    nonexpected = 'fleur.fleur'
    
    assert isinstance(test_and_get_codenode(code, expected), Code)
    with pytest.raises(ValueError) as msg:
        test_and_get_codenode(code, nonexpected, use_exceptions=True)
    assert 'Code not valid' in str(msg)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
Input plug-in for the FLEUR input generator 'inpgen'.
The input generator for the Fleur code is a preprocessor
and should be run localy (with the direct scheduler) or inline,
because it does not take many resources.
"""
from aiida.orm.calculation.job import JobCalculation
from aiida.orm import DataFactory
from aiida.common.exceptions import InputValidationError
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.constants import elements as PeriodicTableElements
from aiida.common.constants import bohr_to_ang
from aiida.common.utils import classproperty
from aiida_fleur.tools.StructureData_util import abs_to_rel_f, abs_to_rel
from aiida_fleur.tools.xml_util import convert_to_fortran_bool, convert_to_fortran_string

StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')

class FleurinputgenCalculation(JobCalculation):
    """
    JobCalculationClass for the Inputgenerator, which is a preprocessor for a FLEUR calculation. Here is implemented how an input file for 'inpgen' is written from AiiDA data objects.
    For more information about produced files and the FLEUR-code family, go to http://www.flapw.de/.
    """
    #### (Maintain) if inputgen keys change ####

    def _init_internal_params(self):
        super(FleurinputgenCalculation, self)._init_internal_params()

        # Default fleur output parser
        self._default_parser = 'fleur.fleurinpgenparser'

        # Default input and output files
        self._DEFAULT_INPUT_FILE = 'aiida.in' # will be shown with inputcat
        self._DEFAULT_OUTPUT_FILE = 'out' #'shell.out' #will be shown with outputcat

        # created file names, some needed for Fleur calc
        self._INPXML_FILE_NAME = 'inp.xml'
        self._INPUT_FILE_NAME = 'aiida.in'
        self._SHELLOUT_FILE_NAME = 'shell.out'
        self._OUTPUT_FILE_NAME = 'out'
        self._ERROR_FILE_NAME = 'out.error'
        self._STRUCT_FILE_NAME = 'struct.xsf'
        '''
        self._INP_FILE_NAME = 'inp'
        self._ENPARA_FILE_NAME = 'enpara'
        self._SYMOUT_FILE_NAME = 'sym.out'
        self._FORT_FILE_NAME = 'fort93'
        self._CORELEVEL_FILE_NAME = 'corelevels.' # Add coordination number
        '''
        self._settings_keys = ['additional_retrieve_list', 'remove_from_retrieve_list',
                               'cmdline']
    # TODO switch all these to init_interal_params?
    _OUTPUT_SUBFOLDER = './fleur_inp_out/'
    _PREFIX = 'aiida'

    # Additional files that should always be retrieved for the specific plugin
    _internal_retrieve_list = []
    _automatic_namelists = {}
    # Specify here what namelist and parameters the inpgen takes
    #TODO: complete?
    _possible_namelists = ['title', 'input', 'lattice', 'gen', 'shift', 'factor', 'qss',
                           'soc', 'atom', 'comp', 'exco', 'film', 'kpt', 'end']
                           # this order is important!
    _possible_params = {'input':['film', 'cartesian', 'cal_symm', 'checkinp',
                                 'symor', 'oldfleur'],
                        'lattice':['latsys', 'a0', 'a', 'b', 'c', 'alpha',
                                   'beta', 'gamma'],
                        'atom':['id', 'z', 'rmt', 'dx', 'jri', 'lmax',
                                'lnonsph', 'ncst', 'econfig', 'bmu', 'lo',
                                'element', 'name'],
                        'comp' : ['jspins', 'frcor', 'ctail', 'kcrel', 'gmax',
                                  'gmaxxc', 'kmax'],
                        'exco' : ['xctyp', 'relxc'],
                        'film' : ['dvac', 'dtild'],
                        'soc' : ['theta', 'phi'],
                        'qss' : ['x', 'y', 'z'],
                        'kpt' : ['nkpt', 'kpts', 'div1', 'div2', 'div3',
                                 'tkb', 'tria'],
                        'title' : {}
                       }

    # Keywords that cannot be set
    # TODO: To specify what combinations are not allowed together,
    # or not at all (lattice ?, shift, scale)
    _blocked_keywords = []

    # TODO different kpt mody? (use kpointNode)? FleurinpdData can do it.
    _use_kpoints = False

    # If two lattices are given, via the input &lattice
    # and the aiida structure prefare the aiida structure?
    # currently is not allow the use of &lattice
    _use_aiida_structure = False

    # Default title
    _inp_title = 'A Fleur input generator calulation with aiida'


    @classproperty
    def _use_methods(cls):
        """
        Extend the parent _use_methods with further keys.
        specifies what nodes have to be in calculation TODO: decide what is
        settings and what is parameters, sturcture might not be needed
        if &lattice is defined in inp
        """
        retdict = JobCalculation._use_methods
        retdict.update({
            "structure": {
                'valid_types': StructureData,
                'additional_parameter': None,
                'linkname': 'structure',
                'docstring': "Choose the input structure to use",
                },
            "parameters": {
                'valid_types': ParameterData,
                'additional_parameter': None,
                'linkname': 'parameters',
                'docstring': ("Use a node that specifies the input parameters "
                              "for the namelists"),
                },
            "settings": {
                'valid_types': ParameterData,
                'additional_parameter': None,
                'linkname': 'settings',
                'docstring': (
                    "This parameter data node is used to specify for some "
                    "advanced features how the plugin behaves. You can add files"
                    "the retrieve list, or add command line switches, "
                    "for all available features here check the documentation."),
            }})
        return retdict


    def _prepare_for_submission(self, tempfolder, inputdict):
        """
        This is the routine to be called when you want to create
        the input files for the inpgen with the plug-in.

        :param tempfolder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputdata_dict (without the Code!)
        """

        #from aiida.common.utils import get_unique_filename, get_suggestion
        #import re

        # Get the connection between coordination number and element symbol
        # maybe do in a differnt way
        _atomic_numbers = {data['symbol']: num for num,
                           data in PeriodicTableElements.iteritems()}

        possible_namelists = self._possible_namelists
        possible_params = self._possible_params
        local_copy_list = []
        remote_copy_list = []
        remote_symlink_list = []
        bulk = True
        film = False

        # convert these 'booleans' to the inpgen format.
        replacer_values_bool = [True, False, 'True', 'False', 't', 'T',
                                'F', 'f']
        # some keywords require a string " around them in the input file.
        string_replace = ['econfig', 'lo', 'element', 'name']

        # of some keys only the values are writen to the file, specify them here.
        val_only_namelist = ['soc', 'qss']

        # Scaling comes from the Structure
        # but we have to convert from Angstroem to a.u (bohr radii)
        scaling_factors = [1.0, 1.0, 1.0] #
        scaling_lat = 1.#/bohr_to_ang
        scaling_pos = 1./bohr_to_ang # Angstrom to atomic
        own_lattice = False #not self._use_aiida_structure

        # The inpfile gen is run in serial TODO: How to do this by default?
        #self.set_withmpi(False)

        ##########################################
        ############# INPUT CHECK ################
        ##########################################

        # first check existence of structure and if 1D, 2D, 3D
        try:
            structure = inputdict.pop(self.get_linkname('structure'))
        except KeyError:
            raise InputValidationError("No structure specified for this"
                                       " calculation")
        if not isinstance(structure, StructureData):
            raise InputValidationError(
                       "structure is not of type StructureData")

        pbc = structure.pbc
        if False in pbc:
            bulk = False
            film = True

        # check existence of parameters (optional)
        parameters = inputdict.pop(self.get_linkname('parameters'), None)
        if parameters is None:
            # use default
            parameters_dict = {}
        else:
            if not isinstance(parameters, ParameterData):
                raise InputValidationError("parameters, if specified, must be of "
                                           "type ParameterData")
            parameters_dict = _lowercase_dict(parameters.get_dict(),
                                              dict_name='parameters')

        # we write always out rel coordinates, because thats the way FLEUR uses
        # them best. we have to convert them from abs, becauses thats how they
        #are stored in a Structure node. cartesian=F is default
        if 'input' in parameters_dict:
            parameters_dict['input']['cartesian'] = False
            if film:
                parameters_dict['input']['film'] = True
        else:
            if bulk:
                parameters_dict['input'] = {'cartesian' : False}
            elif film:
                parameters_dict['input'] = {'cartesian' : False, 'film' : True}

        namelists_toprint = possible_namelists

        # check parameters keys TODO: values needed, or keep plug-in as stupid as possible?
        #if parameters_dict:# TODO remove, unnesseary now?
        input_params = parameters_dict
        #TODO:?make everything lowercase in the database, and change it to inpgen format?
        #_lowercase_dict(parameters.get_dict(),
                                         #dict_name='parameters')
        #input_params = {k: _lowercase_dict(val, dict_name=k)
        #                   for k, val in input_params.iteritems()}
        #input_params_keys = input_params.keys()

        if 'title' in input_params.keys():
            self._inp_title = input_params.pop('title')
        #TODO validate type of values of the input parameter keys ?

        #check input_parameters
        for namelist, paramdic in input_params.iteritems():
            if 'atom' in namelist: # this namelist can be specified more often
                # special atom namelist needs to be set for writing,
                #  but insert it in the right spot!
                index = namelists_toprint.index('atom') + 1
                namelists_toprint.insert(index, namelist)
                namelist = 'atom'
            if namelist not in possible_namelists:
                raise InputValidationError(
                    "The namelist '{}' is not supported by the fleur"
                    " inputgenerator. Check on the fleur website or add '{}'"
                    "to _possible_namelists.".format(namelist, namelist))
            for para in paramdic.keys():
                if para not in possible_params[namelist]:
                    raise InputValidationError(
                        "The property '{}' is not supported by the "
                        "namelist '{}'. "
                        "Check the fleur website, or if it really is,"
                        " update _possible_params. ".format(para, namelist))
                if para in string_replace:
                    #TODO check if its in the parameter dict
                    paramdic[para] = convert_to_fortran_string(paramdic[para])
                # things that are in string replace can never be a bool
                # Otherwise input where someone given the title 'F' would fail...
                elif paramdic[para] in replacer_values_bool:
                    # because 1/1.0 == True, and 0/0.0 == False
                    # maybe change in convert_to_fortran that no error occurs
                    if isinstance(paramdic[para], (int, float)):
                        if isinstance(paramdic[para], bool):
                            paramdic[para] = convert_to_fortran_bool(paramdic[para])
                    else:
                        paramdic[para] = convert_to_fortran_bool(paramdic[para])


            #in fleur it is possible to give a lattice namelist
            if 'lattice' in input_params.keys():
                own_lattice = True
                structure = inputdict.pop(self.get_linkname('structure'), None)
                if structure is not None: #two structures given?
                    #which one should be prepared? TODO: log warning or even error
                    if self._use_aiida_structure:
                        if not isinstance(structure, StructureData):
                            raise InputValidationError("structure is not of type"
                                                       " StructureData")
                        input_params.pop('lattice', {})
                        own_lattice = False




        '''
        # TODO allow only usual kpt meshes and use therefore Aiida kpointData
        if self._use_kpoints:
            try:
                kpoints = inputdict.pop(self.get_linkname('kpoints'))
            except KeyError:
                raise InputValidationError("No kpoints specified for this"
                                           " calculation")
            if not isinstance(kpoints, KpointsData):
                raise InputValidationError("kpoints is not of type KpointsData")
        '''

        #TODO I think the code should not be in the input dict. check local,
        # verus remote, codeinfos, one several codes..
        try:
            code = inputdict.pop(self.get_linkname('code'))
        except KeyError:
            raise InputValidationError("No code specified for this "
                                       "calculation")

        # check existence of settings (optional)
        settings = inputdict.pop(self.get_linkname('settings'), None)
        #print('settings: {}'.format(settings))
        if settings is None:
            settings_dict = {}
        else:
            if not isinstance(settings, ParameterData):
                raise InputValidationError("settings, if specified, must be of "
                                           "type ParameterData")
            else:
                settings_dict = settings.get_dict()
        #check for for allowed keys, ignor unknown keys but warn.
        for key in settings_dict.keys():
            if key not in self._settings_keys:
                #TODO warrning
                self.logger.info("settings dict key {} for Fleur calculation"
                                 "not reconized, only {} are allowed."
                                 "".format(key, self._settings_keys))

        # Here, there should be no more parameters...
        if inputdict:
            raise InputValidationError(
                "The following input data nodes are "
                "unrecognized: {}".format(inputdict.keys()))

        ##############################
        # END OF INITIAL INPUT CHECK #


        #
        #######################################################
        ######### PREPARE PARAMETERS FOR INPUT FILE ###########

        #### STRUCTURE_PARAMETERS ####

        scaling_factor_card = ""
        cell_parameters_card = ""

        if not own_lattice:
            cell = structure.cell
            for vector in cell:
                scaled = [a*scaling_pos for a  in vector]#scaling_pos=1./bohr_to_ang
                cell_parameters_card += ("{0:18.10f} {1:18.10f} {2:18.10f}"
                                         "\n".format(scaled[0], scaled[1], scaled[2]))
            scaling_factor_card += ("{0:18.10f} {1:18.10f} {2:18.10f}"
                                    "\n".format(scaling_factors[0],
                                                scaling_factors[1],
                                                scaling_factors[2]))


        #### ATOMIC_POSITIONS ####

        # TODO: be careful with units
        atomic_positions_card_list = [""]
        atomic_positions_card_listtmp = [""]
        # Fleur does not have any keyword before the atomic species.
        # first the number of atoms then the form nuclear charge, postion
        # Fleur hast the option of nuclear charge as floats,
        # allows the user to distinguish two atoms and break the symmetry.
        if not own_lattice:
            natoms = len(structure.sites)

            #for FLEUR true, general not, because you could put several
            # atoms on a site
            # TODO: test that only one atom at site?

            # TODO this feature might change in Fleur, do different. that in inpgen kind gets a name, which will also be the name in fleur inp.xml.
            # now user has to make kind_name = atom id.
            for site in structure.sites:
                kind_name = site.kind_name
                kind = structure.get_kind(kind_name)
                if kind.has_vacancies():
                    # then we do not at atoms with weights smaller one
                    if kind.weights[0] <1.0:
                        natoms = natoms -1
                        # Log message?
                        continue
                site_symbol = kind.symbols[0] # TODO: list I assume atoms therefore I just get the first one...
                atomic_number = _atomic_numbers[site_symbol]
                atomic_number_name = atomic_number
                if site_symbol != kind_name: # This is an important fact, if usere renames it becomes a new species!
                    suc = True
                    try:
                        head = kind_name.rstrip('0123456789')
                        kind_namet = int(kind_name[len(head):])
                    except ValueError:
                        suc = False
                    if suc:
                        atomic_number_name = '{}.{}'.format(atomic_number, kind_namet)
                # per default we use relative coordinates in Fleur
                # we have to scale back to atomic units from angstrom
                pos = site.position

                if bulk:
                    vector_rel = abs_to_rel(pos, cell)
                elif film:
                    vector_rel = abs_to_rel_f(pos, cell, structure.pbc)
                    vector_rel[2] = vector_rel[2]*scaling_pos
                atomic_positions_card_listtmp.append(
                    "    {0:3} {1:18.10f} {2:18.10f} {3:18.10f}"
                    "\n".format(atomic_number_name,
                                vector_rel[0], vector_rel[1],
                                vector_rel[2]))
                #TODO check format
            # we write it later, since we do not know what natoms is before the loop...
            atomic_positions_card_list.append("    {0:3}\n".format(natoms))
            for card in atomic_positions_card_listtmp:
                atomic_positions_card_list.append(card)
        else:
            # TODO with own lattice atomic positions have to come from somewhere
            # else.... User input?
            raise InputValidationError("fleur lattice needs also the atom "
                                       " position as input,"
                                       " not implemented yet, sorry!")
        atomic_positions_card = "".join(atomic_positions_card_list)
        del atomic_positions_card_list # Free memory

        #### Kpts ####

        # TODO: kpts
        #kpoints_card = ""#.join(kpoints_card_list)
        #del kpoints_card_list

        #######################################
        #### WRITE ALL CARDS IN INPUT FILE ####

        input_filename = tempfolder.get_abs_path(self._INPUT_FILE_NAME)

        with open(input_filename, 'w') as infile:

            #first write title
            infile.write("{0}\n".format(self._inp_title))

            #then write &input namelist
            infile.write("&{0}".format('input'))

            # namelist content; set to {} if not present, so that we leave an
            # empty namelist
            namelist = input_params.pop('input', {})
            for k, val in sorted(namelist.iteritems()):
                infile.write(get_input_data_text(k, val, False, mapping=None))
            infile.write("/\n")

            # Write lattice information now
            infile.write(cell_parameters_card)
            infile.write("{0:18.10f}\n".format(scaling_lat))
            infile.write(scaling_factor_card)
            infile.write("\n")

            # Write Atomic positons
            infile.write(atomic_positions_card)

            # Write namelists after atomic positions
            for namels_name in namelists_toprint:
                namelist = input_params.pop(namels_name, {})
                if namelist:
                    if 'atom' in namels_name:
                        namels_name = 'atom'
                    infile.write("&{0}\n".format(namels_name))
                    if namels_name in val_only_namelist:
                        for k, val in sorted(namelist.iteritems()):
                            infile.write(get_input_data_text(k, val, True, mapping=None))
                    else:
                        for k, val in sorted(namelist.iteritems()):
                            infile.write(get_input_data_text(k, val, False, mapping=None))
                    infile.write("/\n")
            #infile.write(kpoints_card)

        if input_params:
            raise InputValidationError(
                "input_params leftover: The following namelists are specified"
                " in input_params, but are "
                "not valid namelists for the current type of calculation: "
                "{}".format(",".join(input_params.keys())))


        calcinfo = CalcInfo()

        calcinfo.uuid = self.uuid

        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.remote_symlink_list = remote_symlink_list

        # Retrieve per default only out file and inp.xml file?
        retrieve_list = []

        # TODO: let the user specify?
        #settings_retrieve_list = settings_dict.pop(
        #                             'ADDITIONAL_RETRIEVE_LIST', [])
        retrieve_list.append(self._INPXML_FILE_NAME)
        retrieve_list.append(self._OUTPUT_FILE_NAME)
        retrieve_list.append(self._SHELLOUT_FILE_NAME)
        retrieve_list.append(self._ERROR_FILE_NAME)
        retrieve_list.append(self._STRUCT_FILE_NAME)
        retrieve_list.append(self._INPUT_FILE_NAME)
        #calcinfo.retrieve_list += settings_retrieve_list
        #calcinfo.retrieve_list += self._internal_retrieve_list

        # user specific retrieve
        add_retrieve = settings_dict.get('additional_retrieve_list', [])
        #print('add_retrieve: {}'.format(add_retrieve))
        for file1 in add_retrieve:
            retrieve_list.append(file1)

        remove_retrieve = settings_dict.get('remove_from_retrieve_list', [])
        for file1 in remove_retrieve:
            if file1 in retrieve_list:
                retrieve_list.remove(file1)

        calcinfo.retrieve_list = []
        for file1 in retrieve_list:
            calcinfo.retrieve_list.append(file1)

        codeinfo = CodeInfo()
        cmdline_params = ["-explicit"]#, "-electronConfig"] # TODO? let the user decide -electronconfig?
        # econfig write out of defaults is still buggy... but for core level workflows it is quite essential

        # user specific commandline_options
        for command in settings_dict.get('cmdline', []):
            cmdline_params.append(command)
        codeinfo.cmdline_params = (list(cmdline_params))

        codeinfo.code_uuid = code.uuid
        codeinfo.stdin_name = self._INPUT_FILE_NAME
        codeinfo.stdout_name = self._SHELLOUT_FILE_NAME # shell output will be piped in file
        codeinfo.stderr_name = self._ERROR_FILE_NAME # std error too

        calcinfo.codes_info = [codeinfo]

        '''
        if settings_dict:
            try:
                Parserclass = self.get_parserclass()
                parser = Parserclass(self)
                parser_opts = parser.get_parser_settings_key()
                settings_dict.pop(parser_opts)
            except (KeyError, AttributeError): # the key parser_opts isn't
                                              # inside the dictionary
                raise InputValidationError(
                    "The following keys have been found in the settings "
                    "input node, but were not understood: {}"
                    "".format(",".join(settings_dict.keys())))
        '''
        return calcinfo


def write_inpgen_inputfile_aiida_nodes(structure, parameterdata):
    pass


def write_inpgen_inputfile(structure_dict, parameter_dict):
    """
    routine to write an input file without any aiia nodes...
    """
    pass



def get_input_data_text(key, val, value_only, mapping=None):#TODO rewrite for fleur/ delete unnessesariy parts
    """
    Given a key and a value, return a string (possibly multiline for arrays)
    with the text to be added to the input file.

    :param key: the flag name
    :param val: the flag value. If it is an array, a line for each element
            is produced, with variable indexing starting from 1.
            Each value is formatted using the conv_to_fortran function.
    :param mapping: Optional parameter, must be provided if val is a dictionary.
            It maps each key of the 'val' dictionary to the corresponding
            list index. For instance, if ``key='magn'``,
            ``val = {'Fe': 0.1, 'O': 0.2}`` and ``mapping = {'Fe': 2, 'O': 1}``,
            this function will return the two lines ``magn(1) = 0.2`` and
            ``magn(2) = 0.1``. This parameter is ignored if 'val'
            is not a dictionary.
    """
    from aiida.common.utils import conv_to_fortran
    # I don't try to do iterator=iter(val) and catch TypeError because
    # it would also match strings
    # I check first the dictionary, because it would also matc
    # hasattr(__iter__)
    if isinstance(val, dict):
        if mapping is None:
            raise ValueError("If 'val' is a dictionary, you must provide also "
                             "the 'mapping' parameter")

        # At difference with the case of a list, at the beginning
        # list_of_strings
        # is a list of 2-tuples where the first element is the idx, and the
        # second is the actual line. This is used at the end to
        # resort everything.
        list_of_strings = []
        for elemk, itemval in val.iteritems():
            try:
                idx = mapping[elemk]
            except KeyError:
                raise ValueError("Unable to find the key '{}' in the mapping "
                                 "dictionary".format(elemk))

            list_of_strings.append((
                idx, "  {0}({2})={1} ".format(key, conv_to_fortran(itemval), idx)))
                #changed {0}({2}) = {1}\n".format

        # I first have to resort, then to remove the index from the first
        # column, finally to join the strings
        list_of_strings = zip(*sorted(list_of_strings))[1]
        return "".join(list_of_strings)
    elif hasattr(val, '__iter__'):
        if value_only:
            list_of_strings = [
                "  ({1}){0} ".format(conv_to_fortran(itemval), idx+1)
                for idx, itemval in enumerate(val)]
        else:
            # a list/array/tuple of values
            list_of_strings = [
                "  {0}({2})={1} ".format(key, conv_to_fortran(itemval),
                                         idx+1)
                for idx, itemval in enumerate(val)]
        return "".join(list_of_strings)
    else:
        # single value
        #return "  {0}={1} ".format(key, conv_to_fortran(val))
        if value_only:
            return " {0} ".format(val)
        else:
            return "  {0}={1} ".format(key, val)

# TODO import from QE?
def _lowercase_dict(dic, dict_name):
    """
    converts every entry in a dictionary to lowercase
    :param dic: parameters dictionary
    :param dict_name: dictionary name
    """
    from collections import Counter

    if isinstance(dic, dict):
        new_dict = dict((str(k).lower(), val) for k, val in dic.iteritems())
        if len(new_dict) != len(dic):
            num_items = Counter(str(k).lower() for k in dic.keys())
            double_keys = ",".join([k for k, val in num_items if val > 1])
            raise InputValidationError(
                "Inside the dictionary '{}' there are the following keys that "
                "are repeated more than once when compared case-insensitively:"
                "{}.This is not allowed.".format(dict_name, double_keys))
        return new_dict
    else:
        raise TypeError("_lowercase_dict accepts only dictionaries as argument")




# export_extras
#def test_export_extras():
#    pass

# import_extras

# create_group
def test_create_group():
   from aiida_fleur.tools.common_aiida import create_group
   from aiida.orm import DataFactory, Group
   ParameterData = DataFactory('parameter')

   para = ParameterData(dict={})
   # para.store()
   group = create_group([para.pk], 'test_group')
    
   assert isinstance(group, Group)
   # assert 


# get_nodes_from_group
#def test_get_nodes_from_group_uuidlist():
#    pass




import pytest
from aiida.utils.fixtures import fixture_manager

@pytest.fixture(scope='session')
def aiida_env():
    with fixture_manager() as manager:
        yield manager

@pytest.fixture()
def fresh_aiida_env(aiida_env):
    yield
    aiida_env.reset_db()

def test_my_stuff(fresh_aiida_env):
   # run a test
   print('test_my_stuf works')



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

# import dummy_wc from where ever it is
from aiida_fleur.workflows.dummy import dummy_wc
from pprint import pprint
from aiida.orm.data.base import Str
from aiida.work.run import submit

input_s = Str('hello world!')


res = dummy_wc.run(str_display=input_s)

# if you check the output nodes of the submit run, there will be none
res1 = submit(dummy_wc, str_display=input_s)

#res2 = async(dummy_wc, str_display=input_s)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
Here we collect io routines and their utility. For writting certain things to files.
For example collection of data or database evaluations, for other people.
"""

def write_results_to_file(headerstring, data, destination='./outputfile', seperator='  ', transpose=True):
    """
    Writes data to a file

    param headerstring: string with information
    param data: 2D array (numpy,scipy) with data [colum1 colum2, ...]
    """

    thefile = open(destination, 'w')
    thefile.write(headerstring)
    datastring = ''
    seperator = seperator# '\t'
    if transpose:
        datat = data.transpose()
    else:
        datat = data
    for item in datat:
        itemstring = ''
        for value in item:
            if isinstance(value, str) or isinstance(value, basestring):
                itemstring = itemstring + '{}{}'.format(value, seperator)
            else:
                itemstring = itemstring + '{0:0.8f}{1:s}'.format(float(value), seperator)
        datastring = datastring + itemstring.strip() + '\n'
    thefile.write(datastring)
    thefile.close()



def write_xps_spectra_datafile(nodes, factors, natomtypes_dict, bindingenergies, bindingenergies_ref, xdata_spec, ydata_spec, ydata_single_all, xdata_all, ydata_all, compound_info, xdatalabel, destination='./outputfile'):
    '''
    special file write routine. Writes theoretical spectra data from plot spectra to file

    uses write_results_to_file
    '''
    import numpy as np

    dataprep = [xdata_spec, ydata_spec] + ydata_single_all
    data = np.array(dataprep)#, ydata_single_all])

    formulastring = compound_info, factors
    nodesinvolvedstring = nodes
    atomtypesstring = natomtypes_dict
    be_string = bindingenergies
    reference_be_string = bindingenergies_ref

    headstring = ('# Theoretical Data provided as is without warranty. Copyright 2017-2018 Forschungszentrum Juelich GmbH\n'
                  '# produced at PGI-1 with the FLEUR code within the AiiDA framework (with aiida-fleur), MIT License, please cite: \n'
                  '################# Data meta information #############\n')
    tempst = ('# System: {}\n# Nodes: {}\n# Elements and natomtypes: {}\n# Bindingenergies [eV], pos: {}\n# Reference Energies [eV] pos: {}\n'
              ''.format(formulastring, nodesinvolvedstring, atomtypesstring, be_string, reference_be_string))
    tempst1 = ''
    for label in xdatalabel:
        tempst1 = tempst1 + ' | ' + label
    tempst2 = ('#####################  Data  ######################\n# Energy [eV] | Total intensity {}\n'.format(tempst1))

    headstring =  headstring + tempst + tempst2

    print('Writting theoretical XPS data to file: {}'.format(destination))
    write_results_to_file(headstring, data, destination=destination, seperator='  ')

# example/test
#from plot_methods.plot_fleur_aiida import plot_spectra
#import plot_methods
#all_wc_BeTi_uuid = ['107b0727-15cf-4436-b614-79801cdadd8c', 'f8b12b23-0b71-45a1-9040-b51ccf379439']
#factors = [1,1]
#returnvalues = plot_spectra(all_wc_BeTi_uuid, factors=factors, energy_range=[109,112], energy_grid=0.2)
#write_xps_spectra_datafile(all_wc_BeTi_uuid, factors, *returnvalues, destination='./out.txt')





import pytest


# read-cif_folder
@pytest.mark.usefixtures("aiida_env")
def test_read_cif_folder_interface(aiida_env):
    """
    this test set reads in the cif files in the ../files/cif/ directory and subdirs
    it stores the datastructures, with some comments and extras.

    """
    import os
    from aiida_fleur.tools.read_cif_folder import read_cif_folder
    
    # preparation
    
    cif_folderpath = os.path.abspath('../files/cif/')
    print cif_folderpath
    #read_in
    structure_data, filenames = read_cif_folder(path=os.getcwd(), rekursive=True,
                                                store=False, log=False,
                                               comments='', extras='')
    #test number of structurs written
    #test number of cif files written
    #test if extras are set right
    #test prov
    #test
    pass

    #read_in again
    # test if cif files are not rewritten.



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This file is just were to hardcode some schema file paths
"""

import os

# any additional schema file add here, plugin will find them
schema_file_paths = ['./input/0.27/FleurInputSchema.xsd', './input/0.28/FleurInputSchema.xsd', '.']


package_directory = os.path.dirname(os.path.abspath(__file__))


def get_schema_paths():
    """
    returns all know hardcodes schemas as a list of abs paths
    """
    schema_paths = []
    for schema in schema_file_paths:
        path = os.path.abspath(os.path.join(package_directory, schema))
        if os.path.isfile(path):
           schema_paths.append(path)
    return schema_paths

def get_internal_search_paths():
    """
    returns all abs paths to dirs where schema files might be
    """
    #schema_paths = []
    #for schema in schema_file_paths:
    #    path = os.path.abspath(os.path.join(package_directory, schema))
    #    if os.path.isdir(path):
    #       schema_paths.append(path)
    schema_paths = [package_directory]
    return schema_paths



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

import pytest

@pytest.mark.usefixtures("aiida_env")
class TestAiida_fleur_entrypoints:
    """
    tests all the entry points of the Fleur plugin. Therefore if the plugin is 
    reconized by AiiDA and installed right. 
    """
    
    # Calculations
    
    def test_inpgen_calculation_entry_point(aiida_env):
        from aiida.orm import CalculationFactory
        inpgen_calculation = CalculationFactory('fleur.inpgen')
        assert inpgen_calculation is not None

    def test_fleur_calculation_entry_point(aiida_env):
        from aiida.orm import CalculationFactory
        
        fleur_calculation = CalculationFactory('fleur.fleur')
        assert fleur_calculation is not None
    
    
    # Data
    
    def test_fleur_fleurinpdata_entry_point(aiida_env):
        from aiida.orm import DataFactory
        from aiida_fleur.data.fleurinp import FleurinpData
    
        fleurinp = DataFactory('fleur.fleurinp')
        assert fleurinp == FleurinpData

    def test_fleur_fleurinpmodifier_entry_point(aiida_env):
        from aiida.orm import DataFactory
        from aiida_fleur.data.fleurinpmodifier import FleurinpModifier

        fleurinpmod = DataFactory('fleur.fleurinpmodifier')
        assert fleurinpmod == FleurinpModifier


    # Parsers

    def test_inpgen_parser_entry_point(aiida_env):
        from aiida.parsers import ParserFactory
        from aiida_fleur.parsers.fleur_inputgen import Fleur_inputgenParser

        parser = ParserFactory('fleur.fleurinpgenparser')
        assert parser == Fleur_inputgenParser

    def test_fleur_parser_entry_point(aiida_env):
        from aiida.parsers import ParserFactory
        from aiida_fleur.parsers.fleur import FleurParser

        parser = ParserFactory('fleur.fleurparser')
        assert parser == FleurParser


    # Workflows/workchains

    def test_fleur_scf_wc_entry_point(aiida_env):
        from aiida.orm import WorkflowFactory
        from aiida_fleur.workflows.scf import fleur_scf_wc
        
        workflow = WorkflowFactory('fleur.scf')
        assert workflow == fleur_scf_wc
        
    def test_fleur_dos_wc_entry_point(aiida_env):
        from aiida.orm import WorkflowFactory
        from aiida_fleur.workflows.dos import fleur_dos_wc
        
        workflow = WorkflowFactory('fleur.dos')
        assert workflow == fleur_dos_wc
        
    def test_fleur_band_wc_entry_point(aiida_env):
        from aiida.orm import WorkflowFactory
        from aiida_fleur.workflows.band import fleur_band_wc
        
        workflow = WorkflowFactory('fleur.band')
        assert workflow == fleur_band_wc

    def test_fleur_eos_wc_entry_point(aiida_env):
        from aiida.orm import WorkflowFactory
        from aiida_fleur.workflows.eos import fleur_eos_wc
        
        workflow = WorkflowFactory('fleur.eos')
        assert workflow == fleur_eos_wc
    
    # this entry point has currently a problem...
    def test_fleur_inital_cls_wc_entry_point(aiida_env):
        from aiida.orm import WorkflowFactory
        from aiida_fleur.workflows.initial_cls import fleur_initial_cls_wc

        workflow = WorkflowFactory('fleur.init_cls')
        assert workflow == fleur_initial_cls_wc
        
    def test_fleur_corehole_wc_entry_point(aiida_env):
        from aiida.orm import WorkflowFactory
        from aiida_fleur.workflows.corehole import fleur_corehole_wc
        
        workflow = WorkflowFactory('fleur.corehole')
        assert workflow == fleur_corehole_wc
